"""
Taken from Caduceus repository: https://github.com/kuleshov-group/caduceus/blob/main/caduceus/modeling_caduceus.py
Allows for easier loading without using AutoModel from HuggingFace.
"""

import inspect
import math
from functools import partial
from typing import Optional, Tuple, Union

import torch
from mamba_ssm.modules.mamba_simple import Mamba
from mamba_ssm.modules.block import Block  # mambav2 file structure
from torch import nn
from torch.nn import functional as F
from transformers import PreTrainedModel
from transformers.modeling_outputs import BaseModelOutputWithNoAttention


from mamba_ssm.ops.triton.layer_norm import (
    RMSNorm,
    layer_norm_fn,
    rms_norm_fn,
)  # mambav2 file structure
from .pretrained_config import CaduceusPretrainedConfig
from .modeling_rcps import RCPSAddNormWrapper, RCPSEmbedding, RCPSMambaBlock


def create_block(
    d_model,
    ssm_cfg=None,
    norm_epsilon=1e-5,
    rms_norm=False,
    residual_in_fp32=False,
    fused_add_norm=False,
    layer_idx=None,
    bidirectional=True,
    bidirectional_strategy="add",
    bidirectional_weight_tie=True,
    rcps=False,
    device=None,
    dtype=None,
):
    """Create Caduceus block.

    Adapted from: https://github.com/state-spaces/mamba/blob/main/mamba_ssm/models/mixer_seq_simple.py
    """
    if ssm_cfg is None:
        ssm_cfg = {}
    factory_kwargs = {"device": device, "dtype": dtype}
    bidirectional_kwargs = {
        "bidirectional": bidirectional,
        "bidirectional_strategy": bidirectional_strategy,
        "bidirectional_weight_tie": bidirectional_weight_tie,
    }
    mixer_cls = partial(
        BiMambaWrapper,
        layer_idx=layer_idx,
        **ssm_cfg,
        **bidirectional_kwargs,
        **factory_kwargs,
    )
    norm_cls = partial(
        nn.LayerNorm if not rms_norm else RMSNorm, eps=norm_epsilon, **factory_kwargs
    )
    block_cls = RCPSMambaBlock if rcps else Block
    # mambav2 compatibility
    if "mlp_cls" in inspect.signature(block_cls.__init__).parameters:
        block = block_cls(
            d_model,
            mixer_cls,
            mlp_cls=nn.Identity,
            norm_cls=norm_cls,
            fused_add_norm=fused_add_norm,
            residual_in_fp32=residual_in_fp32,
        )
    else:
        block = block_cls(
            d_model,
            mixer_cls,
            norm_cls=norm_cls,
            fused_add_norm=fused_add_norm,
            residual_in_fp32=residual_in_fp32,
        )
    block.layer_idx = layer_idx
    return block


class BiMambaWrapper(nn.Module):
    """Thin wrapper around Mamba to support bi-directionality."""

    def __init__(
        self,
        d_model: int,
        bidirectional: bool = True,
        bidirectional_strategy: Optional[str] = "add",
        bidirectional_weight_tie: bool = True,
        **mamba_kwargs,
    ):
        super().__init__()
        if bidirectional and bidirectional_strategy is None:
            bidirectional_strategy = "add"  # Default strategy: `add`
        if bidirectional and bidirectional_strategy not in ["add", "ew_multiply"]:
            raise NotImplementedError(
                f"`{bidirectional_strategy}` strategy for bi-directionality is not implemented!"
            )
        self.bidirectional = bidirectional
        self.bidirectional_strategy = bidirectional_strategy
        self.mamba_fwd = Mamba(d_model=d_model, **mamba_kwargs)
        if bidirectional:
            self.mamba_rev = Mamba(d_model=d_model, **mamba_kwargs)
            if (
                bidirectional_weight_tie
            ):  # Tie in and out projections (where most of param count lies)
                self.mamba_rev.in_proj.weight = self.mamba_fwd.in_proj.weight
                self.mamba_rev.in_proj.bias = self.mamba_fwd.in_proj.bias
                self.mamba_rev.out_proj.weight = self.mamba_fwd.out_proj.weight
                self.mamba_rev.out_proj.bias = self.mamba_fwd.out_proj.bias
        else:
            self.mamba_rev = None

    def forward(self, hidden_states, inference_params=None):
        """Bidirectional-enabled forward pass

        hidden_states: (B, L, D)
        Returns: same shape as hidden_states
        """
        out = self.mamba_fwd(hidden_states, inference_params=inference_params)
        if self.bidirectional:
            out_rev = self.mamba_rev(
                hidden_states.flip(
                    dims=(1,)
                ),  # Flip along the sequence length dimension
                inference_params=inference_params,
            ).flip(
                dims=(1,)
            )  # Flip back for combining with forward hidden states
            if self.bidirectional_strategy == "add":
                out = out + out_rev
            elif self.bidirectional_strategy == "ew_multiply":
                out = out * out_rev
            else:
                raise NotImplementedError(
                    f"`{self.bidirectional_strategy}` for bi-directionality not implemented!"
                )
        return out


class CaduceusEmbeddings(nn.Module):
    def __init__(
        self,
        config: CaduceusPretrainedConfig,
        device=None,
        dtype=None,
    ):
        super().__init__()
        factory_kwargs = {"device": device, "dtype": dtype}
        if config.rcps:
            self.word_embeddings = RCPSEmbedding(
                config.vocab_size,
                config.d_model,
                config.complement_map,
                **factory_kwargs,
            )
        else:
            self.word_embeddings = nn.Embedding(
                config.vocab_size, config.d_model, **factory_kwargs
            )

    def forward(self, input_ids):
        """
        input_ids: (batch, seqlen)
        """
        return self.word_embeddings(input_ids)


class CaduceusMixerModel(nn.Module):
    def __init__(
        self,
        config: CaduceusPretrainedConfig,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        factory_kwargs = {"device": device, "dtype": dtype}

        self.fused_add_norm = config.fused_add_norm
        self.rcps = config.rcps
        self.residual_in_fp32 = config.residual_in_fp32

        self.embeddings = CaduceusEmbeddings(config, **factory_kwargs)

        # Mamba changes the order of residual and layer norm:
        # Instead of LN -> Attn / MLP -> Add, we do:
        # Add -> LN -> Attn / MLP / Mixer, returning both the residual branch (output of Add) and
        # the main branch (output of MLP / Mixer). The model definition is unchanged.
        # This is for performance reason: we can fuse add + layer_norm.
        if config.fused_add_norm:
            if layer_norm_fn is None or rms_norm_fn is None:
                raise ImportError("Failed to import Triton LayerNorm / RMSNorm kernels")

        self.layers = nn.ModuleList(
            [
                create_block(
                    config.d_model,
                    ssm_cfg=config.ssm_cfg,
                    norm_epsilon=config.norm_epsilon,
                    rms_norm=config.rms_norm,
                    residual_in_fp32=config.residual_in_fp32,
                    fused_add_norm=config.fused_add_norm,
                    layer_idx=i,
                    bidirectional=config.bidirectional,
                    bidirectional_strategy=config.bidirectional_strategy,
                    bidirectional_weight_tie=config.bidirectional_weight_tie,
                    rcps=config.rcps,
                    **factory_kwargs,
                )
                for i in range(config.n_layer)
            ]
        )

        norm_f = (nn.LayerNorm if not config.rms_norm else RMSNorm)(
            config.d_model, eps=config.norm_epsilon, **factory_kwargs
        )
        self.norm_f = (
            norm_f
            if (config.fused_add_norm or not config.rcps)
            else RCPSAddNormWrapper(norm_f)
        )

    def forward(self, input_ids, inputs_embeds=None, output_hidden_states=False):
        """Mixer forward."""
        all_hidden_states = []
        if inputs_embeds is not None:
            hidden_states = inputs_embeds
        else:
            hidden_states = self.embeddings(input_ids)

        residual = None
        for layer in self.layers:
            if output_hidden_states:
                all_hidden_states.append(hidden_states)
            # TODO: Add support for gradient checkpointing
            hidden_states, residual = layer(
                hidden_states, residual, inference_params=None
            )

        if not self.fused_add_norm:
            if self.rcps:
                # Set prenorm=False here since we don't need the residual
                hidden_states = self.norm_f(
                    hidden_states, residual=residual, prenorm=False
                )
            else:
                residual = (
                    (hidden_states + residual)
                    if residual is not None
                    else hidden_states
                )
                hidden_states = self.norm_f(residual.to(dtype=self.norm_f.weight.dtype))
        else:
            fused_add_norm_fn = (
                rms_norm_fn if isinstance(self.norm_f, RMSNorm) else layer_norm_fn
            )
            if self.rcps:
                # Set prenorm=False here since we don't need the residual
                hidden_states_fwd = fused_add_norm_fn(
                    hidden_states[..., : hidden_states.shape[-1] // 2],
                    self.norm_f.weight,
                    self.norm_f.bias,
                    eps=self.norm_f.eps,
                    residual=residual[..., : hidden_states.shape[-1] // 2],
                    prenorm=False,
                    residual_in_fp32=self.residual_in_fp32,
                )
                hidden_states_rc = fused_add_norm_fn(
                    hidden_states[..., hidden_states.shape[-1] // 2 :].flip(
                        dims=[-2, -1]
                    ),
                    self.norm_f.weight,
                    self.norm_f.bias,
                    eps=self.norm_f.eps,
                    residual=residual[..., hidden_states.shape[-1] // 2 :].flip(
                        dims=[-2, -1]
                    ),
                    prenorm=False,
                    residual_in_fp32=self.residual_in_fp32,
                )
                hidden_states = torch.cat(
                    [hidden_states_fwd, hidden_states_rc.flip(dims=[-2, -1])], dim=-1
                )
            else:
                # Set prenorm=False here since we don't need the residual
                hidden_states = fused_add_norm_fn(
                    hidden_states,
                    self.norm_f.weight,
                    self.norm_f.bias,
                    eps=self.norm_f.eps,
                    residual=residual,
                    prenorm=False,
                    residual_in_fp32=self.residual_in_fp32,
                )
            if output_hidden_states:
                all_hidden_states.append(hidden_states)
        return hidden_states, all_hidden_states


class CaduceusPreTrainedModel(PreTrainedModel):
    """PreTrainedModel wrapper for Caduceus backbone."""

    config_class = CaduceusPretrainedConfig
    base_model_prefix = "caduceus"
    supports_gradient_checkpointing = False
    _no_split_modules = ["BiMambaWrapper"]

    def _init_weights(
        self,
        module,
        initializer_range=0.02,  # Now only used for embedding layer.
        **kwargs,
    ):
        """Adapted from: https://github.com/state-spaces/mamba/blob/main/mamba_ssm/models/mixer_seq_simple.py"""

        n_layer = self.config.n_layer
        initialized_cfg = (
            self.config.initializer_cfg
            if self.config.initializer_cfg is not None
            else {}
        )
        rescale_prenorm_residual = initialized_cfg.get("rescale_prenorm_residual", True)
        initializer_range = initialized_cfg.get("initializer_range", initializer_range)
        n_residuals_per_layer = initialized_cfg.get("n_residuals_per_layer", 1)

        if isinstance(module, nn.Linear):
            if module.bias is not None:
                if not getattr(module.bias, "_no_reinit", False):
                    nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Embedding):
            nn.init.normal_(module.weight, std=initializer_range)

        if rescale_prenorm_residual:
            # Reinitialize selected weights subject to the OpenAI GPT-2 Paper Scheme:
            #   > A modified initialization which accounts for the accumulation on the residual path with model depth.
            #   > Scale the weights of residual layers at initialization by a factor of 1/√N where N is the # of
            #   residual layers.
            #   >   -- GPT-2 :: https://openai.com/blog/better-language-models/
            #
            # Reference (Megatron-LM): https://github.com/NVIDIA/Megatron-LM/blob/main/megatron/model/gpt_model.py
            for name, p in module.named_parameters():
                if name in ["out_proj.weight", "fc2.weight"]:
                    # Special Scaled Initialization --> There are 2 Layer Norms per Transformer Block
                    # Following Pytorch init, except scale by 1/sqrt(2 * n_layer)
                    # We need to reinit p since this code could be called multiple times
                    # Having just p *= scale would repeatedly scale it down
                    nn.init.kaiming_uniform_(p, a=math.sqrt(5))
                    with torch.no_grad():
                        p /= math.sqrt(n_residuals_per_layer * n_layer)


class CaduceusModel(CaduceusPreTrainedModel):
    """Caduceus model that can be instantiated using HF patterns."""

    def __init__(
        self, config: CaduceusPretrainedConfig, device=None, dtype=None, **kwargs
    ):
        super().__init__(config)

        if config.rcps:
            assert (
                config.complement_map is not None
            ), "Complement map must be provided for RCPS."

        # Adjust vocab size and complement maps if vocab padding is set.
        if config.vocab_size % config.pad_vocab_size_multiple != 0:
            config.vocab_size += config.pad_vocab_size_multiple - (
                config.vocab_size % config.pad_vocab_size_multiple
            )
        if config.complement_map is not None and config.vocab_size > len(
            config.complement_map
        ):
            for i in range(len(config.complement_map), config.vocab_size):
                config.complement_map[i] = i

        self.config = config
        factory_kwargs = {"device": device, "dtype": dtype}
        self.backbone = CaduceusMixerModel(config, **factory_kwargs, **kwargs)

    def forward(
        self,
        input_ids: torch.LongTensor = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[torch.Tensor, Tuple, BaseModelOutputWithNoAttention]:
        """HF-compatible forward method."""
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        hidden_states, all_hidden_states = self.backbone(
            input_ids,
            inputs_embeds=inputs_embeds,
            output_hidden_states=output_hidden_states,
        )
        if return_dict:
            return BaseModelOutputWithNoAttention(
                last_hidden_state=hidden_states,
                hidden_states=all_hidden_states if output_hidden_states else None,
            )
        elif output_hidden_states:
            return hidden_states, all_hidden_states
        else:
            return hidden_states



from helical.models.caduceus import CaduceusFineTuningModel, CaduceusConfig
import hydra
from omegaconf import DictConfig


@hydra.main(
    version_base=None,
    config_path="../run_models/configs",
    config_name="caduceus_config",
)
def run_fine_tuning(cfg: DictConfig):
    input_sequences = ["ACT" * 20, "ATG" * 10, "ATG" * 20, "CTG" * 10, "TTG" * 20]
    labels = [0, 2, 2, 0, 1]

    caduceus_config = CaduceusConfig(**cfg)
    caduceus_fine_tune = CaduceusFineTuningModel(
        caduceus_config=caduceus_config,
        fine_tuning_head="classification",
        output_size=3,
    )

    train_dataset = caduceus_fine_tune.process_data(input_sequences)

    caduceus_fine_tune.train(train_dataset=train_dataset, train_labels=labels)

    outputs = caduceus_fine_tune.get_outputs(train_dataset)
    print(outputs)

    # save and load model
    caduceus_fine_tune.save_model("./caduceus_fine_tuned_model.pt")
    caduceus_fine_tune.load_model("./caduceus_fine_tuned_model.pt")

    outputs = caduceus_fine_tune.get_outputs(train_dataset)
    print(outputs)  

if __name__ == "__main__":
    run_fine_tuning()



from enum import Enum


class LoggingType(str, Enum):
    FILE = "FILE"
    CONSOLE = "CONSOLE"
    FILE_AND_CONSOLE = "FILE_AND_CONSOLE"
    NOTSET = "NOTSET"


class LoggingLevel(str, Enum):
    CRITICAL = "CRITICAL"
    FATAL = "FATAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    WARN = "WARN"
    INFO = "INFO"
    DEBUG = "DEBUG"
    NOTSET = "NOTSET"



import pickle as pkl
import requests
import json
import pickle as pkl

# imports
import logging
import torch
from tqdm.auto import trange
import re
import torch
from transformers import BertForMaskedLM
import pandas as pd
import numpy as np


logger = logging.getLogger(__name__)


def load_mappings(gene_symbols):
    server = "https://rest.ensembl.org"
    ext = "/lookup/symbol/homo_sapiens"
    headers = {"Content-Type": "application/json", "Accept": "application/json"}

    # r = requests.post(server+ext, headers=headers, data='{ "symbols" : ["BRCA2", "BRAF" ] }')

    gene_id_to_ensemble = dict()

    for i in range(0, len(gene_symbols), 1000):
        # print(i+1000,"/",len(test.var['gene_symbols']))
        symbols = {"symbols": gene_symbols[i : i + 1000].tolist()}
        r = requests.post(server + ext, headers=headers, data=json.dumps(symbols))
        decoded = r.json()
        gene_id_to_ensemble.update(decoded)
        # print(repr(decoded))

    pkl.dump(gene_id_to_ensemble, open("./human_gene_to_ensemble_id.pkl", "wb"))
    return gene_id_to_ensemble


def _compute_embeddings_depending_on_mode(
    embeddings: torch.tensor,
    data_dict: dict,
    emb_mode: str,
    cls_present: bool,
    eos_present: bool,
    token_to_ensembl_dict: dict,
):
    """
    Compute the different embeddings for each emb_mode

    Parameters
    -----------
    embeddings: torch.tensor
        The embedding batch output by the model.
    data_dict: dict
        The minibatch data dictionary used an input to the model.
    emb_mode: str
        The mode in which the embeddings are to be computed.
    cls_present: bool
        Whether the <cls> token is present in the token dictionary.
    eos_present: bool
        Whether the <eos> token is present in the token dictionary.
    token_to_ensembl_dict: dict
        The token to ensemble dictionary from the .
    """
    if emb_mode == "cell":
        length = data_dict["length"]
        if cls_present:
            embeddings = embeddings[:, 1:, :]  # Get all layers except the cls embs
            if eos_present:
                length -= 2  # length is used for the mean calculation, 2 is subtracted because we have taken both the cls and eos embeddings out
            else:
                length -= 1  # length is subtracted because just the cls is removed

        batch_embeddings = mean_nonpadding_embs(embeddings, length).cpu().numpy()

    elif emb_mode == "gene":
        if cls_present:
            embeddings = embeddings[:, 1:, :]
            if eos_present:
                embeddings = embeddings[:, :-1, :]

        batch_embeddings = []
        for embedding, ids in zip(embeddings, data_dict["input_ids"]):
            cell_dict = {}
            if cls_present:
                ids = ids[1:]
                if eos_present:
                    ids = ids[:-1]
            for id, gene_emb in zip(ids, embedding):
                cell_dict[token_to_ensembl_dict[id.item()]] = gene_emb.cpu().numpy()

            batch_embeddings.append(pd.Series(cell_dict))

    elif emb_mode == "cls":
        batch_embeddings = embeddings[:, 0, :]  # CLS token layer

    return batch_embeddings


def _check_for_expected_special_tokens(
    dataset, emb_mode, cls_present, eos_present, gene_token_dict
):
    """
    Check for the expected special tokens in the dataset.

    Parameters
    -----------
    dataset: dict
        The batch dictionary with input ids.
    emb_mode: str
        The mode in which the embeddings are to be computed.
    cls_present: bool
        Whether the <cls> token is present in the token dictionary.
    eos_present: bool
        Whether the <eos> token is present in the token dictionary.
    gene_token_dict: dict
        The gene token dictionary from the tokenizer.
    """
    if emb_mode == "cls":
        message = "<cls> token missing in token dictionary"
        if not cls_present:
            logger.error(message)
            raise ValueError(message)

        if dataset["input_ids"][0][0] != gene_token_dict["<cls>"]:
            message = "First token is not <cls> token value"
            logger.error(message)
            raise ValueError(message)

    elif emb_mode == "cell":
        if cls_present:
            logger.warning(
                "CLS token present in token dictionary, excluding from average."
            )
        if eos_present:
            logger.warning(
                "EOS token present in token dictionary, excluding from average."
            )


# extract embeddings
def get_embs(
    model,
    filtered_input_data,
    emb_mode,
    layer_to_quant,
    pad_token_id,
    forward_batch_size,
    gene_token_dict,
    token_to_ensembl_dict,
    cls_present,
    eos_present,
    device,
    silent=False,
    output_attentions=False,
    output_genes=False,
):
    model_input_size = get_model_input_size(model)
    total_batch_length = len(filtered_input_data)
    embs_list = []
    attn_list = []
    input_genes = []

    _check_for_expected_special_tokens(
        filtered_input_data, emb_mode, cls_present, eos_present, gene_token_dict
    )

    overall_max_len = 0
    with torch.no_grad(), torch.amp.autocast("cuda", enabled=True):
        for i in trange(0, total_batch_length, forward_batch_size, leave=(not silent)):
            max_range = min(i + forward_batch_size, total_batch_length)

            minibatch = filtered_input_data.select([i for i in range(i, max_range)])

            minibatch.set_format(type="torch", device=device)
            lengths = minibatch["length"]
            max_len = int(max(lengths))

            input_data_minibatch = minibatch["input_ids"]
            if output_genes:
                for input_ids in input_data_minibatch:
                    input_ids = input_ids.cpu().numpy()
                    gene_list = []
                    for id in input_ids:
                        gene_list.append(token_to_ensembl_dict[id])
                    input_genes.append(gene_list)
            input_data_minibatch = pad_tensor_list(
                input_data_minibatch, max_len, pad_token_id, model_input_size
            ).to(device)

            outputs = model(
                input_ids=input_data_minibatch,
                attention_mask=gen_attention_mask(minibatch),
                output_attentions=output_attentions,
            )

            embs_i = outputs.hidden_states[layer_to_quant]
            # attention of size (batch_size, num_heads, sequence_length, sequence_length)
            if output_attentions:
                attn_i = outputs.attentions[layer_to_quant]
                # attn_i = torch.mean(attn_i, dim=1).cpu().numpy()  # average over heads
                attn_list.extend(attn_i.cpu().numpy())

            embs_list.extend(
                _compute_embeddings_depending_on_mode(
                    embs_i,
                    {
                        "input_ids": input_data_minibatch,
                        "length": lengths,
                    },
                    emb_mode,
                    cls_present,
                    eos_present,
                    token_to_ensembl_dict,
                )
            )
            overall_max_len = max(overall_max_len, max_len)
            del outputs
            del minibatch
            del input_data_minibatch
            del embs_i

            # torch.cuda.empty_cache()
    if emb_mode != "gene":
        embs_list = np.array(embs_list)

    if output_attentions:
        if output_genes:
            return embs_list, attn_list, input_genes
        return embs_list, attn_list

    if output_genes:
        return embs_list, input_genes

    return embs_list


def downsample_and_sort(data, max_ncells):
    num_cells = len(data)
    # if max number of cells is defined, then shuffle and subsample to this max number
    if max_ncells is not None:
        if num_cells > max_ncells:
            data = data.shuffle(seed=42)
            num_cells = max_ncells
    data_subset = data.select([i for i in range(num_cells)])
    # sort dataset with largest cell first to encounter any memory errors earlier
    data_sorted = data_subset.sort("length", reverse=True)
    return data_sorted


def quant_layers(model):
    layer_nums = []
    for name, parameter in model.named_parameters():
        if "layer" in name:
            layer_nums += [int(name.split("layer.")[1].split(".")[0])]
    return int(max(layer_nums)) + 1


def get_model_input_size(model):
    return int(re.split("\(|,", str(model.bert.embeddings.position_embeddings))[1])


def load_model(model_type, model_directory, device):
    if model_type == "Pretrained":
        model = BertForMaskedLM.from_pretrained(
            model_directory, output_hidden_states=True, output_attentions=False
        )
    # put the model in eval mode for fwd pass and load onto the GPU if available
    model.eval()
    model = model.to(device)
    return model


def pad_tensor(tensor, pad_token_id, max_len):
    tensor = torch.nn.functional.pad(
        tensor, pad=(0, max_len - tensor.numel()), mode="constant", value=pad_token_id
    )

    return tensor


def pad_3d_tensor(tensor, pad_token_id, max_len, dim):
    if dim == 0:
        raise Exception("dim 0 usually does not need to be padded.")
    if dim == 1:
        pad = (0, 0, 0, max_len - tensor.size()[dim])
    elif dim == 2:
        pad = (0, max_len - tensor.size()[dim], 0, 0)
    tensor = torch.nn.functional.pad(
        tensor, pad=pad, mode="constant", value=pad_token_id
    )
    return tensor


# pad list of tensors and convert to tensor
def pad_tensor_list(
    tensor_list,
    dynamic_or_constant,
    pad_token_id,
    model_input_size,
    dim=None,
    padding_func=None,
):
    # determine maximum tensor length
    if dynamic_or_constant == "dynamic":
        max_len = max([tensor.squeeze().numel() for tensor in tensor_list])
    elif isinstance(dynamic_or_constant, int):
        max_len = dynamic_or_constant
    else:
        max_len = model_input_size
        logger.warning(
            "If padding style is constant, must provide integer value. "
            f"Setting padding to max input size {model_input_size}."
        )

    # pad all tensors to maximum length
    if dim is None:
        tensor_list = [
            pad_tensor(tensor, pad_token_id, max_len) for tensor in tensor_list
        ]
    else:
        tensor_list = [
            padding_func(tensor, pad_token_id, max_len, dim) for tensor in tensor_list
        ]
    # return stacked tensors
    if padding_func != pad_3d_tensor:
        return torch.stack(tensor_list)
    else:
        return torch.cat(tensor_list, 0)


def gen_attention_mask(minibatch_encoding, max_len=None):
    if max_len is None:
        max_len = max(minibatch_encoding["length"])
    original_lens = minibatch_encoding["length"]
    attention_mask = [
        (
            [1] * original_len + [0] * (max_len - original_len)
            if original_len <= max_len
            else [1] * max_len
        )
        for original_len in original_lens
    ]
    return torch.tensor(attention_mask, device=minibatch_encoding["length"].device)


# get cell embeddings excluding padding
def mean_nonpadding_embs(embs, original_lens, dim=1):
    # create a mask tensor based on padding lengths
    mask = torch.arange(embs.size(dim), device=embs.device) < original_lens.unsqueeze(1)
    if embs.dim() == 3:
        # fill the masked positions in embs with zeros
        masked_embs = embs.masked_fill(~mask.unsqueeze(2), 0.0)

        # compute the mean across the non-padding dimensions
        mean_embs = masked_embs.sum(dim) / original_lens.view(-1, 1).float()

    elif embs.dim() == 2:
        masked_embs = embs.masked_fill(~mask, 0.0)
        mean_embs = masked_embs.sum(dim) / original_lens.float()
    return mean_embs



from helical.models.uce import UCEConfig
import pytest


@pytest.mark.parametrize(
    "model_name, n_layers", [("33l_8ep_1024t_1280", 33), ("4layer_model", 4)]
)
def test_uce_config_valid_model_names(model_name, n_layers):
    """
    Test case for the UCE config initialization.

    Args:
        model_name (str): The name of the model.
        n_layers (int): The number of layers of the model.
    """
    configurer = UCEConfig(model_name=model_name)
    assert configurer.config["model_path"].name == f"{model_name}.torch"
    assert configurer.config["n_layers"] == n_layers


@pytest.mark.parametrize("model_name", ["wrong_name"])
def test_uce_config__invalid_model_names(model_name):
    """
    Test case when an invalid model name is provided.
    Verifies that a ValueError is raised when an invalid model name is passed to the UCEConfig constructor.

    Parameters:
    - model_name (str): The invalid model name.

    Raises:
    - ValueError: If the model name is invalid.
    """
    with pytest.raises(ValueError):
        UCEConfig(model_name=model_name)



"""Character tokenizer for Hugging Face."""

from typing import List, Optional, Dict, Tuple

from transformers import PreTrainedTokenizer


class CharTokenizer(PreTrainedTokenizer):
    model_input_names = ["input_ids"]

    def __init__(
        self,
        model_max_length: int,
        bos_token="[BOS]",
        eos_token="[SEP]",
        sep_token="[SEP]",
        cls_token="[CLS]",
        pad_token="[PAD]",
        mask_token="[MASK]",
        unk_token="[UNK]",
        **kwargs,
    ):
        """Character tokenizer for Hugging Face transformers.
        Adapted from https://huggingface.co/LongSafari/hyenadna-tiny-1k-seqlen-hf/blob/main/tokenization_hyena.py
        Args:
            model_max_length (int): Model maximum sequence length.
            characters (Sequence[str]): List of desired characters. Any character which
                is not included in this list will be replaced by a special token called
                [UNK] with id=6. Following is a list of the special tokens with
                their corresponding ids:
                    "[CLS]": 0
                    "[SEP]": 1
                    "[BOS]": 2
                    "[MASK]": 3
                    "[PAD]": 4
                    "[RESERVED]": 5
                    "[UNK]": 6
                an id (starting at 7) will be assigned to each character.
        """
        self.characters = ("A", "C", "G", "U", "N")
        self.model_max_length = model_max_length

        self._vocab_str_to_int = {
            "[CLS]": 0,
            "[SEP]": 1,
            "[BOS]": 2,
            "[MASK]": 3,
            "[PAD]": 4,
            "[RESERVED]": 5,
            "[UNK]": 6,
            **{ch: i + 7 for i, ch in enumerate(self.characters)},
        }

        self._vocab_int_to_str = {v: k for k, v in self._vocab_str_to_int.items()}
        add_prefix_space = kwargs.pop("add_prefix_space", False)
        padding_side = kwargs.pop("padding_side", "left")

        self._vocab_str_to_int["T"] = self._vocab_str_to_int["U"]

        super().__init__(
            bos_token=bos_token,
            eos_token=eos_token,
            sep_token=sep_token,
            cls_token=cls_token,
            pad_token=pad_token,
            mask_token=mask_token,
            unk_token=unk_token,
            add_prefix_space=add_prefix_space,
            model_max_length=model_max_length,
            padding_side=padding_side,
            **kwargs,
        )

    @property
    def vocab_size(self) -> int:
        return len(self._vocab_str_to_int)

    def _tokenize(self, text: str, **kwargs) -> List[str]:
        return list(text.upper())  # Convert all base pairs to uppercase

    def _convert_token_to_id(self, token: str) -> int:
        return self._vocab_str_to_int.get(token, self._vocab_str_to_int["[UNK]"])

    def _convert_id_to_token(self, index: int) -> str:
        return self._vocab_int_to_str[index]

    def convert_tokens_to_string(self, tokens):
        return "".join(
            tokens
        )  # Note: this operation has lost info about which base pairs were originally lowercase

    def get_special_tokens_mask(
        self,
        token_ids_0: List[int],
        token_ids_1: Optional[List[int]] = None,
        already_has_special_tokens: bool = False,
    ) -> List[int]:
        if already_has_special_tokens:
            return super().get_special_tokens_mask(
                token_ids_0=token_ids_0,
                token_ids_1=token_ids_1,
                already_has_special_tokens=True,
            )

        result = ([0] * len(token_ids_0)) + [1]
        if token_ids_1 is not None:
            result += ([0] * len(token_ids_1)) + [1]
        return result

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        sep = [self.sep_token_id]
        # cls = [self.cls_token_id]
        result = token_ids_0 + sep
        if token_ids_1 is not None:
            result += token_ids_1 + sep
        return result

    def get_vocab(self) -> Dict[str, int]:
        return self._vocab_str_to_int

    # Fixed vocabulary with no vocab file
    def save_vocabulary(
        self, save_directory: str, filename_prefix: Optional[str] = None
    ) -> Tuple:
        return ()



from helical.utils.downloader import Downloader
from pathlib import Path
import logging

LOGGER = logging.getLogger(__name__)


def download_geneformer_models():
    downloader = Downloader()
    versions = ["v1", "v2"]

    # We can decide to download more models by simply adding the model names from the full list as reported in geneformer_config.py
    version_models_dict = {
        "v1": ["gf-12L-40M-i2048", "gf-6L-10M-i2048"],
        "v2": ["gf-12L-38M-i4096", "gf-12L-38M-i4096-CLcancer", "gf-20L-151M-i4096"],
    }

    for version in versions:
        # Download common files for each version
        common_files = [
            f"geneformer/{version}/gene_median_dictionary.pkl",
            f"geneformer/{version}/token_dictionary.pkl",
            f"geneformer/{version}/ensembl_mapping_dict.pkl",
        ]
        for file in common_files:
            downloader.download_via_name(file)

        # Get all model directories
        model_dirs = version_models_dict[version]

        for model_name in model_dirs:
            model_files = [
                f"geneformer/{version}/{model_name}/config.json",
                f"geneformer/{version}/{model_name}/training_args.bin",
            ]

            # Add version-specific files
            if version == "v2":
                model_files.extend(
                    [
                        f"geneformer/{version}/{model_name}/generation_config.json",
                        f"geneformer/{version}/{model_name}/model.safetensors",
                    ]
                )
            else:
                model_files.append(
                    f"geneformer/{version}/{model_name}/pytorch_model.bin"
                )

            # Download all files for the current model
            for file in model_files:
                downloader.download_via_name(file)

    LOGGER.info("All Geneformer models and files have been downloaded.")


def main():
    downloader = Downloader()
    downloader.display = False

    downloader.download_via_name("uce/4layer_model.torch")
    downloader.download_via_name("uce/all_tokens.torch")
    downloader.download_via_name("uce/species_chrom.csv")
    downloader.download_via_name("uce/species_offsets.pkl")
    downloader.download_via_name(
        "uce/protein_embeddings/Homo_sapiens.GRCh38.gene_symbol_to_embedding_ESM2.pt"
    )
    downloader.download_via_name(
        "uce/protein_embeddings/Macaca_fascicularis.Macaca_fascicularis_6.0.gene_symbol_to_embedding_ESM2.pt"
    )
    downloader.download_via_name(
        "uce/protein_embeddings/Danio_rerio.GRCz11.gene_symbol_to_embedding_ESM2.pt"
    )
    downloader.download_via_name(
        "uce/protein_embeddings/Macaca_mulatta.Mmul_10.gene_symbol_to_embedding_ESM2.pt"
    )
    downloader.download_via_name(
        "uce/protein_embeddings/Microcebus_murinus.Mmur_3.0.gene_symbol_to_embedding_ESM2.pt"
    )
    downloader.download_via_name(
        "uce/protein_embeddings/Mus_musculus.GRCm39.gene_symbol_to_embedding_ESM2.pt"
    )
    downloader.download_via_name(
        "uce/protein_embeddings/Sus_scrofa.Sscrofa11.1.gene_symbol_to_embedding_ESM2.pt"
    )
    downloader.download_via_name(
        "uce/protein_embeddings/Xenopus_tropicalis.Xenopus_tropicalis_v9.1.gene_symbol_to_embedding_ESM2.pt"
    )

    downloader.download_via_name("scgpt/scGPT_CP/vocab.json")
    downloader.download_via_name("scgpt/scGPT_CP/best_model.pt")

    download_geneformer_models()

    downloader.download_via_name("hyena_dna/hyenadna-tiny-1k-seqlen.ckpt")
    downloader.download_via_name("hyena_dna/hyenadna-tiny-1k-seqlen-d256.ckpt")
    downloader.download_via_name("hyena_dna/hyenadna-small-32k-seqlen.ckpt")
    downloader.download_via_name("hyena_dna/hyenadna-medium-450k-seqlen.ckpt")
    downloader.download_via_name("hyena_dna/hyenadna-large-1m-seqlen.ckpt")

    downloader.download_via_name(
        "caduceus/caduceus-ph-16L-seqlen-131k-d256/model.safetensors"
    )
    downloader.download_via_name(
        "caduceus/caduceus-ph-16L-seqlen-131k-d256/config.json"
    )
    downloader.download_via_name(
        "caduceus/caduceus-ph-4L-seqlen-1k-d118/model.safetensors"
    )
    downloader.download_via_name("caduceus/caduceus-ph-4L-seqlen-1k-d118/config.json")
    downloader.download_via_name(
        "caduceus/caduceus-ph-4L-seqlen-1k-d256/model.safetensors"
    )
    downloader.download_via_name("caduceus/caduceus-ph-4L-seqlen-1k-d256/config.json")
    downloader.download_via_name(
        "caduceus/caduceus-ps-16L-seqlen-131k-d256/model.safetensors"
    )
    downloader.download_via_name(
        "caduceus/caduceus-ps-16L-seqlen-131k-d256/config.json"
    )
    downloader.download_via_name(
        "caduceus/caduceus-ps-4L-seqlen-1k-d118/model.safetensors"
    )
    downloader.download_via_name("caduceus/caduceus-ps-4L-seqlen-1k-d118/config.json")
    downloader.download_via_name(
        "caduceus/caduceus-ps-4L-seqlen-1k-d256/model.safetensors"
    )
    downloader.download_via_name("caduceus/caduceus-ps-4L-seqlen-1k-d256/config.json")

    downloader.download_via_name("genept/genept_embeddings/genept_embeddings.json")

    downloader.download_via_link(
        Path("yolksac_human.h5ad"),
        "https://huggingface.co/datasets/helical-ai/yolksac_human/resolve/main/data/17_04_24_YolkSacRaw_F158_WE_annots.h5ad?download=true",
    )
    return True


if __name__ == "__main__":
    main()



"""
Geneformer tokenizer.

**Input data:**

| *Required format:* raw counts scRNAseq data without feature selection as anndata file.
| *Required row (gene) attribute:* "ensembl_id"; Ensembl ID for each gene.
| *Required col (cell) attribute:* "total_counts"; total read counts in that cell.

| *Optional col (cell) attribute:* "filter_pass"; binary indicator of whether cell should be tokenized based on user-defined filtering criteria.
| *Optional col (cell) attributes:* any other cell metadata can be passed on to the tokenized dataset as a custom attribute dictionary as shown below.

**Usage:**

.. code-block :: python

    >>> from geneformer import TranscriptomeTokenizer
    >>> tk = TranscriptomeTokenizer({"cell_type": "cell_type", "organ_major": "organ"}, nproc=4)
    >>> tk.tokenize_data("data_directory", "output_directory", "output_prefix")

**Description:**

| Input data is a directory with .h5ad files containing raw counts from single cell RNAseq data, including all genes detected in the transcriptome without feature selection.
  The input file type is specified by the argument file_format in the tokenize_data function.

| Genes should be labeled with Ensembl IDs, which provide a unique identifer for conversion to tokens. Other forms of gene annotations (e.g. gene names) can be converted to Ensembl IDs via Ensembl Biomart.
  Cells should be labeled with the total read count in the cell to be used for normalization.

| No cell metadata is required, but custom cell attributes may be passed onto the tokenized dataset by providing a dictionary of custom attributes to be added.
  For example, if the original .h5ad dataset has column attributes "cell_type" and "organ_major" and one would like to retain these attributes as labels in the tokenized dataset with the new names "cell_type" and "organ", respectively,
  the following custom attribute dictionary should be provided: {"cell_type": "cell_type", "organ_major": "organ"}.

| Additionally, if the original .h5ad file contains a cell column attribute called "filter_pass", this column will be used as a binary indicator of whether to include these cells in the tokenized data.
  All cells with "1" in this attribute will be tokenized, whereas the others will be excluded. One may use this column to indicate QC filtering or other criteria for selection for inclusion in the final tokenized dataset.

| If one's data is in other formats besides .h5ad, one can use the relevant tools (such as Anndata tools) to convert the file to a .h5ad format prior to running the transcriptome tokenizer.

| OF NOTE: Take care that the correct token dictionary and gene median file is used for the correct model.

| OF NOTE: For 95M model series, special_token should be True and model_input_size should be 4096. For 30M model series, special_token should be False and model_input_size should be 2048.

"""

from __future__ import annotations

import os
import logging
import pickle
import warnings
from pathlib import Path
from typing import Literal
from collections import Counter
import numpy as np
import scipy.sparse as sp
from datasets import Dataset
from anndata._core.views import ArrayView
import pandas as pd
from tqdm import tqdm
import scanpy as sc

warnings.filterwarnings("ignore", message=".*The 'nopython' keyword.*")
LOGGER = logging.getLogger(__name__)


def rank_genes(gene_vector, gene_tokens):
    """
    Rank gene expression vector.
    """
    # sort by median-scaled gene values
    sorted_indices = np.argsort(-gene_vector)
    return gene_tokens[sorted_indices]


def tokenize_cell(gene_vector, gene_tokens):
    """
    Convert normalized gene expression vector to tokenized rank value encoding.
    """
    # create array of gene vector with token indices
    # mask undetected genes
    nonzero_mask = np.nonzero(gene_vector)[0]
    # rank by median-scaled gene values
    return rank_genes(gene_vector[nonzero_mask], gene_tokens[nonzero_mask])


def sum_ensembl_ids(
    data,
    collapse_gene_ids,
    gene_mapping_dict,
    gene_token_dict,
    file_format="h5ad",
    chunk_size=512,
):
    """
    Process genomic data by mapping and potentially collapsing Ensembl IDs.

    This function handles only H5AD file formats. It maps Ensembl IDs
    using a provided dictionary and, if necessary, collapses and sums counts for
    duplicate IDs. If no duplicate IDs are found and collapsing is not necessary, the function
      returns the original data without modifications.
    """
    if file_format == "h5ad":
        """
        Map Ensembl IDs from gene mapping dictionary. If duplicate Ensembl IDs are found, sum counts together.
        Returns adata object with deduplicated Ensembl IDs.
        """

        assert (
            "ensembl_id" in data.var.columns
        ), "'ensembl_id' column missing from data.var"

        # Get the ensembl ids that exist in data
        ensembl_ids = data.var.ensembl_id
        # Check for duplicate Ensembl IDs if collapse_gene_ids is False.
        # Comparing to gene_token_dict here, would not perform any mapping steps
        if not collapse_gene_ids:
            ensembl_id_check = [
                gene for gene in ensembl_ids if gene in gene_token_dict.keys()
            ]
            if len(ensembl_id_check) == len(set(ensembl_id_check)):
                return data
            else:
                raise ValueError("Error: data Ensembl IDs non-unique.")

        ## NOTE genes_in_map_dict that comes from the original paper, has both gene names as keys and ENSEMBL IDs as keys!!
        ## This is probably done to simplify the mapping and checks but not clean. Gene Name: ENSEMBL ID and ENSEMBL ID: ENSEMBL ID (same ensembl ID though).

        # Get the genes that exist in the mapping dictionary and the value of those genes
        genes_in_map_dict = [
            gene for gene in ensembl_ids if gene in gene_mapping_dict.keys()
        ]
        vals_from_map_dict = [gene_mapping_dict.get(gene) for gene in genes_in_map_dict]

        # if the genes in the mapping dict and the value of those genes are of the same length,
        # simply return the mapped values
        if len(set(genes_in_map_dict)) == len(set(vals_from_map_dict)):
            data.var["ensembl_id_collapsed"] = data.var.ensembl_id.str.upper().map(
                gene_mapping_dict
            )
            return data
        # Genes need to be collapsed
        else:
            data.var["ensembl_id_collapsed"] = data.var.ensembl_id.str.upper().map(
                gene_mapping_dict
            )
            data.var_names = data.var["ensembl_id_collapsed"]
            data = data[:, ~data.var.index.isna()]
            dup_genes = [
                idx for idx, count in Counter(data.var_names).items() if count > 1
            ]

            num_chunks = int(np.ceil(data.shape[0] / chunk_size))

            processed_genes = []
            for i in tqdm(range(num_chunks)):
                start_idx = i * chunk_size
                end_idx = min((i + 1) * chunk_size, data.shape[0])
                data_chunk = data[start_idx:end_idx, :]

                processed_chunks = []
                for dup_gene in dup_genes:
                    data_dup_gene = data_chunk[:, data_chunk.var_names == dup_gene]
                    df = pd.DataFrame.sparse.from_spmatrix(
                        data_dup_gene.X,
                        index=data_dup_gene.obs_names,
                        columns=data_dup_gene.var_names,
                    )
                    df_sum = pd.DataFrame(df.sum(axis=1))
                    df_sum.columns = [dup_gene]
                    df_sum.index = data_dup_gene.obs.index
                    processed_chunks.append(df_sum)

                processed_chunks = pd.concat(processed_chunks, axis=1)
                processed_genes.append(processed_chunks)
            processed_genes = pd.concat(processed_genes, axis=0)
            var_df = pd.DataFrame({"ensembl_id_collapsed": processed_genes.columns})
            var_df.index = processed_genes.columns
            processed_genes = sc.AnnData(X=processed_genes, obs=data.obs, var=var_df)

            data_dedup = data[:, ~data.var.index.isin(dup_genes)]  # Deduplicated data
            data_dedup = sc.concat([data_dedup, processed_genes], axis=1)
            data_dedup.obs = data.obs
            return data_dedup
    else:
        raise ValueError(
            f"Unsupported file format {file_format}. Currently, only h5ad is supported."
        )


DEFAULT_GENE_MEDIAN_FILE = Path(__file__).parent / "v1" / "gene_median_dictionary.pkl"
DEFAULT_TOKEN_DICTIONARY_FILE = Path(__file__).parent / "v1" / "token_dictionary.pkl"
DEFAULT_ENSEMBL_MAPPING_FILE = (
    Path(__file__).parent / "v1" / "ensembl_mapping_dictionary.pkl"
)


class TranscriptomeTokenizer:
    def __init__(
        self,
        custom_attr_name_dict=None,
        nproc=1,
        chunk_size=512,
        model_input_size=2048,
        special_token=False,
        collapse_gene_ids=True,
        gene_median_file=DEFAULT_GENE_MEDIAN_FILE,
        token_dictionary_file=DEFAULT_TOKEN_DICTIONARY_FILE,
        gene_mapping_file=DEFAULT_ENSEMBL_MAPPING_FILE,
    ):
        """
        Initialize tokenizer.

        **Parameters:**

        custom_attr_name_dict : None, dict
            | Dictionary of custom attributes to be added to the dataset.
            | Keys are the names of the attributes in the h5ad file.
            | Values are the names of the attributes in the dataset.
        nproc : int
            | Number of processes to use for dataset mapping.
        chunk_size : int = 512
            | Chunk size for anndata tokenizer.
        model_input_size : int = 4096
            | Max input size of model to truncate input to.
            | For the 30M model series, should be 2048. For the 95M model series, should be 4096.
        special_token : bool = True
            | Adds CLS token before and EOS token after rank value encoding.
            | For the 30M model series, should be False. For the 95M model series, should be True.
        collapse_gene_ids : bool = True
            | Whether to collapse gene IDs based on gene mapping dictionary.
        gene_median_file : Path
            | Path to pickle file containing dictionary of non-zero median
            | gene expression values across Genecorpus-30M.
        token_dictionary_file : Path
            | Path to pickle file containing token dictionary (Ensembl IDs:token).
        gene_mapping_file : None, Path
            | Path to pickle file containing dictionary for collapsing gene IDs.
        """

        # dictionary of custom attributes {output dataset column name: input .h5ad column name}
        self.custom_attr_name_dict = custom_attr_name_dict

        # number of processes for dataset mapping
        self.nproc = nproc

        # chunk size for anndata tokenizer
        self.chunk_size = chunk_size

        # input size for tokenization
        self.model_input_size = model_input_size

        # add CLS and EOS tokens
        self.special_token = special_token

        # load dictionary of gene normalization factors
        # (non-zero median value of expression across Genecorpus-30M)
        with open(gene_median_file, "rb") as f:
            self.gene_median_dict = pickle.load(f)

        # load token dictionary (Ensembl IDs:token)
        with open(token_dictionary_file, "rb") as f:
            self.gene_token_dict = pickle.load(f)

        # check for special token in gene_token_dict
        if self.special_token:
            if ("<cls>" not in self.gene_token_dict.keys()) and (
                "<eos>" not in self.gene_token_dict.keys()
            ):
                LOGGER.error(
                    "<cls> and <eos> required in gene_token_dict when special_token = True."
                )
                raise

        if not self.special_token:
            if ("<cls>" in self.gene_token_dict.keys()) and (
                "<eos>" in self.gene_token_dict.keys()
            ):
                LOGGER.warning(
                    "<cls> and <eos> are in gene_token_dict but special_token = False. Please note that for 95M model series, special_token should be True."
                )

        # if collapsing duplicate gene IDs
        self.collapse_gene_ids = collapse_gene_ids

        # load gene mappings dictionary (Ensembl IDs:Ensembl ID)
        if gene_mapping_file is not None:
            with open(gene_mapping_file, "rb") as f:
                self.gene_mapping_dict = pickle.load(f)
        else:
            self.gene_mapping_dict = {k: k for k, _ in self.gene_token_dict.items()}

        # gene keys for full vocabulary
        self.gene_keys = list(self.gene_token_dict.keys())

        #  Filter gene mapping dict for items that exist in gene_token_dict
        gene_keys_set = set(self.gene_token_dict.keys())
        self.gene_mapping_dict = {
            k: v for k, v in self.gene_mapping_dict.items() if v in gene_keys_set
        }

        # Maps a token back to the Ensembl ID
        self.token_to_ensembl_dict = {
            value: key for key, value in self.gene_token_dict.items()
        }

        # protein-coding and miRNA gene list dictionary for selecting .h5ad columns for tokenization
        self.genelist_dict = dict(zip(self.gene_keys, [True] * len(self.gene_keys)))

    def tokenize_data(
        self,
        data_directory: Path | str,
        output_directory: Path | str,
        output_prefix: str,
        file_format: Literal["h5ad"] = "h5ad",
        use_generator: bool = False,
    ):
        """
        Tokenize anndata files in data_directory and save as tokenized .dataset in output_directory.

        **Parameters:**

        data_directory : Path
            | Path to directory containing anndata files
        output_directory : Path
            | Path to directory where tokenized data will be saved as .dataset
        output_prefix : str
            | Prefix for output .dataset
        file_format : str
            | Format of input files. Can only be "h5ad" for now.
        use_generator : bool
            | Whether to use generator or dict for tokenization.
        """
        tokenized_cells, cell_metadata = self.tokenize_files(
            Path(data_directory), file_format
        )
        tokenized_dataset = self.create_dataset(
            tokenized_cells,
            cell_metadata,
            use_generator=use_generator,
        )

        output_path = (Path(output_directory) / output_prefix).with_suffix(".dataset")
        tokenized_dataset.save_to_disk(str(output_path))

    def tokenize_files(self, data_directory, file_format: Literal["h5ad"] = "h5ad"):
        """
        Tokenizes anndata files in a specified directory and returns the tokenized cells and their metadata.

        Parameters:
            data_directory (Path): Path to the directory containing anndata files.
            file_format (str): Format of the input files. Currently, only "h5ad" is supported.

        Returns:
            tuple: A tuple containing the list of tokenized cells and their corresponding metadata.
        """

        if file_format not in ["h5ad"]:
            raise ValueError(
                f"Unsupported file format {file_format}. Currently, only h5ad is supported."
            )

        tokenized_cells = []
        if self.custom_attr_name_dict is not None:
            cell_attr = [attr_key for attr_key in self.custom_attr_name_dict.keys()]
            cell_metadata = {
                attr_key: [] for attr_key in self.custom_attr_name_dict.values()
            }

        # loops through directories to tokenize files
        file_found = 0
        # loops through directories to tokenize files
        for file_path in data_directory.glob(f"*.{file_format}"):
            file_found = 1
            LOGGER.info(f"Tokenizing {file_path}")
            raw_adata = sc.read_h5ad(file_path)
            file_tokenized_cells, file_cell_metadata = self.tokenize_anndata(raw_adata)
            tokenized_cells += file_tokenized_cells
            if self.custom_attr_name_dict is not None:
                for k in cell_attr:
                    cell_metadata[self.custom_attr_name_dict[k]] += file_cell_metadata[
                        k
                    ]
            else:
                cell_metadata = None

        if file_found == 0:
            LOGGER.error(
                f"No .{file_format} files found in directory {data_directory}."
            )
            raise
        return tokenized_cells, cell_metadata

    def tokenize_anndata(self, adata_obj, target_sum=10_000):
        """
        Tokenizes AnnData object into a list of tokenized cells and their corresponding metadata.

        Parameters:
            adata_obj (AnnData): The AnnData object to be tokenized.
            target_sum (int, optional): The target sum for normalization. Defaults to 10_000.

        Returns:
            tuple: A tuple containing a list of tokenized cells and a dictionary of cell metadata.
        """
        adata = sum_ensembl_ids(
            adata_obj,
            self.collapse_gene_ids,
            self.gene_mapping_dict,
            self.gene_token_dict,
            file_format="h5ad",
            chunk_size=self.chunk_size,
        )
        if self.custom_attr_name_dict is not None:
            file_cell_metadata = {
                attr_key: [] for attr_key in self.custom_attr_name_dict.keys()
            }

        coding_miRNA_loc = np.where(
            [self.genelist_dict.get(i, False) for i in adata.var["ensembl_id"]]
        )[0]
        norm_factor_vector = np.array(
            [
                self.gene_median_dict[i]
                for i in adata.var["ensembl_id"].iloc[coding_miRNA_loc]
            ]
        )
        coding_miRNA_ids = adata.var["ensembl_id"].iloc[coding_miRNA_loc]
        coding_miRNA_tokens = np.array(
            [self.gene_token_dict[i] for i in coding_miRNA_ids]
        )

        try:
            _ = adata.obs["filter_pass"]
        except KeyError:
            var_exists = False
        else:
            var_exists = True

        if var_exists:
            filter_pass_loc = np.where([i == 1 for i in adata.obs["filter_pass"]])[0]
        elif not var_exists:
            LOGGER.info(
                f"{adata_obj} has no column attribute 'filter_pass'; tokenizing all cells."
            )
            filter_pass_loc = np.array([i for i in range(adata.shape[0])])

        tokenized_cells = []

        for i in range(0, len(filter_pass_loc), self.chunk_size):
            idx = filter_pass_loc[i : i + self.chunk_size]

            # Check if 'n_counts' exists in adata.obs
            if "n_counts" in adata.obs.columns:
                n_counts = adata[idx].obs["n_counts"].values[:, None]
            else:
                # If 'n_counts' doesn't exist, calculate it as the sum of counts for each cell
                if isinstance(adata.X, ArrayView):
                    # If adata.X is an ArrayView the sum returns only
                    # one dimension, while 2 are required for broadcasting below
                    n_counts = adata[idx, :].X.sum(axis=1)[:, None]
                else:
                    n_counts = adata[idx, :].X.sum(axis=1)

            X_view0 = adata[idx, :].X
            X_view = X_view0[:, coding_miRNA_loc]
            
            # Broadcasting: (batch_size, num_genes) / (batch_size, 1) * target_sum / (1, num_genes)
            X_norm = (X_view / np.asarray(n_counts).reshape(-1, 1)) * target_sum / norm_factor_vector.reshape(1, -1)
            X_norm = sp.csr_matrix(X_norm)

            tokenized_cells += [
                rank_genes(X_norm[i].data, coding_miRNA_tokens[X_norm[i].indices])
                for i in range(X_norm.shape[0])
            ]

            # add custom attributes for subview to dict
            if self.custom_attr_name_dict is not None:
                for k in file_cell_metadata.keys():
                    file_cell_metadata[k] += adata[idx].obs[k].tolist()
            else:
                file_cell_metadata = None

        return tokenized_cells, file_cell_metadata

    def create_dataset(
        self,
        tokenized_cells,
        cell_metadata,
        use_generator=False,
        keep_uncropped_input_ids=False,
    ):
        """
        Creates a dataset from tokenized cells and cell metadata.

        Args:
            tokenized_cells: A list of tokenized cells.
            cell_metadata: A dictionary of cell metadata.
            use_generator (bool): Whether to use a generator to create the dataset. Defaults to False.
            keep_uncropped_input_ids (bool): Whether to keep the original uncropped input_ids. Defaults to False.

        Returns:
            A Hugging Face dataset with formatted cell features.
        """
        LOGGER.info("Creating dataset.")

        cls_token = self.gene_token_dict.get("<cls>")
        eos_token = self.gene_token_dict.get("<eos>")

        processed_inputs = []
        lengths = []
        uncropped_inputs = [] if keep_uncropped_input_ids else None
        uncropped_lengths = [] if keep_uncropped_input_ids else None
        for cell in tokenized_cells:
            if keep_uncropped_input_ids:
                uncropped_inputs.append(cell)
                uncropped_lengths.append(len(cell))

            if self.special_token:
                # truncate but keep room for CLS + EOS
                cropped = cell[: self.model_input_size - 2]
                processed = np.concatenate([[cls_token], cropped, [eos_token]])
            else:
                processed = cell[: self.model_input_size]
            processed_inputs.append(processed)
            lengths.append(len(processed))
        # build dataset dict
        dataset_dict = {
            "input_ids": processed_inputs,
            "length": lengths,
        }

        if keep_uncropped_input_ids:
            dataset_dict["input_ids_uncropped"] = uncropped_inputs
            dataset_dict["length_uncropped"] = uncropped_lengths

        if self.custom_attr_name_dict is not None:
            dataset_dict.update(cell_metadata)

        # create dataset
        if use_generator:

            def dict_generator():
                for i in range(len(processed_inputs)):
                    yield {k: dataset_dict[k][i] for k in dataset_dict}

            output_dataset = Dataset.from_generator(dict_generator, num_proc=self.nproc)
        else:
            output_dataset = Dataset.from_dict(dataset_dict)
        return output_dataset



from torch.utils.data import Dataset
import torch


class HelixmRNADataset(Dataset):
    def __init__(self, sequences, tokenizer):
        self.sequences = sequences
        self.tokenizer = tokenizer
        self.max_length = len(max(sequences, key=len)) + 10
        self.labels = None

    def set_labels(self, labels):
        self.labels = labels

    def __len__(self):
        return len(self.sequences)

    def __getitem__(self, idx):
        seqs = self.sequences[idx]

        # Tokenize all sequences in the batch
        encoded = self.tokenizer(
            seqs,
            truncation=True,
            padding="max_length",
            max_length=self.max_length,
            return_special_tokens_mask=True,
            return_tensors="pt",
        )

        # Prepare output dictionary
        output = {
            "input_ids": encoded["input_ids"].squeeze(0),
            "special_tokens_mask": encoded["special_tokens_mask"].squeeze(0),
        }

        # Add labels if they exist
        if self.labels is not None:
            output["labels"] = torch.tensor(self.labels[idx])

        return output



from helical.models.helix_mrna import HelixmRNA, HelixmRNAConfig
import pytest
from pandas import DataFrame
import torch


class TestHelixmRNAModel:
    @pytest.fixture
    def helixmRNA(self):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        config = HelixmRNAConfig(batch_size=1, device=self.device, max_length=100)
        return HelixmRNA(configurer=config)

    @pytest.fixture
    def mock_data(self, helixmRNA):
        input_sequences = [
            "AAAA",
            "CCCC",
            "UUUU" * 30,
            "ACGU" * 100,
            "ACGN",
            "ANNU" * 20,
        ]
        tokenized_sequences = helixmRNA.process_data(input_sequences)
        return tokenized_sequences

    def test_invalid_sequences(self, helixmRNA):
        input_sequences = [
            "AAQA",
            "CCCZ",
            "FJAK" * 30,
            "QWER" * 100,
            "JFHS",
            "OFNW" * 20,
        ]
        with pytest.raises(ValueError, match=r"Invalid RNA sequence:*"):
            helixmRNA.process_data(input_sequences)

    def test_helix_mrna_get_embeddings(self, mock_data, helixmRNA):
        embeddings = helixmRNA.get_embeddings(mock_data)

        assert (
            embeddings is not None
        ), f"Embeddings should not be None for sequence: {mock_data}"

    @pytest.mark.parametrize(
        "data, raise_exception",
        [
            (
                DataFrame(
                    {
                        "Sequence": [
                            "EACU" * 20,
                            "EAUG" * 20,
                            "EAUG" * 20,
                            "EACU" * 20,
                            "EAUU" * 20,
                        ]
                    }
                ),
                False,
            ),
            (
                DataFrame(
                    {
                        "invalid": [
                            "EACU" * 20,
                            "EAUG" * 20,
                            "EAUG" * 20,
                            "EACU" * 20,
                            "EAUU" * 20,
                        ]
                    }
                ),
                True,
            ),
        ],
    )
    def test_process_data_w_dataframe_raises_appropriate_errors(
        self, helixmRNA, data, raise_exception
    ):
        if raise_exception:
            with pytest.raises(
                KeyError, match=r"The DataFrame must have a column named 'Sequence'."
            ):
                helixmRNA.process_data(data)
        else:
            helixmRNA.process_data(data)



"""

BSD 3-Clause License

Copyright (c) Soumith Chintala 2016, 
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Copied from torchvision to remove that dependency
https://github.com/pytorch/vision/blob/release/0.21/torchvision/utils.py
https://github.com/pytorch/vision/blob/release/0.21/torchvision/ops/stochastic_depth.py
"""

import torch
import torch.fx
from torch import nn, Tensor
from typing import Any
from types import FunctionType

def _log_api_usage_once(obj: Any) -> None:

    """
    Logs API usage(module and name) within an organization.
    In a large ecosystem, it's often useful to track the PyTorch and
    TorchVision APIs usage. This API provides the similar functionality to the
    logging module in the Python stdlib. It can be used for debugging purpose
    to log which methods are used and by default it is inactive, unless the user
    manually subscribes a logger via the `SetAPIUsageLogger method <https://github.com/pytorch/pytorch/blob/eb3b9fe719b21fae13c7a7cf3253f970290a573e/c10/util/Logging.cpp#L114>`_.
    Please note it is triggered only once for the same API call within a process.
    It does not collect any data from open-source users since it is no-op by default.
    For more information, please refer to
    * PyTorch note: https://pytorch.org/docs/stable/notes/large_scale_deployments.html#api-usage-logging;
    * Logging policy: https://github.com/pytorch/vision/issues/5052;

    Args:
        obj (class instance or method): an object to extract info from.
    """
    module = obj.__module__
    if not module.startswith("torchvision"):
        module = f"torchvision.internal.{module}"
    name = obj.__class__.__name__
    if isinstance(obj, FunctionType):
        name = obj.__name__
    torch._C._log_api_usage_once(f"{module}.{name}")

def stochastic_depth(input: Tensor, p: float, mode: str, training: bool = True) -> Tensor:
    """
    Implements the Stochastic Depth from `"Deep Networks with Stochastic Depth"
    <https://arxiv.org/abs/1603.09382>`_ used for randomly dropping residual
    branches of residual architectures.

    Args:
        input (Tensor[N, ...]): The input tensor or arbitrary dimensions with the first one
                    being its batch i.e. a batch with ``N`` rows.
        p (float): probability of the input to be zeroed.
        mode (str): ``"batch"`` or ``"row"``.
                    ``"batch"`` randomly zeroes the entire input, ``"row"`` zeroes
                    randomly selected rows from the batch.
        training: apply stochastic depth if is ``True``. Default: ``True``

    Returns:
        Tensor[N, ...]: The randomly zeroed tensor.
    """
    if not torch.jit.is_scripting() and not torch.jit.is_tracing():
        _log_api_usage_once(stochastic_depth)
    if p < 0.0 or p > 1.0:
        raise ValueError(f"drop probability has to be between 0 and 1, but got {p}")
    if mode not in ["batch", "row"]:
        raise ValueError(f"mode has to be either 'batch' or 'row', but got {mode}")
    if not training or p == 0.0:
        return input

    survival_rate = 1.0 - p
    if mode == "row":
        size = [input.shape[0]] + [1] * (input.ndim - 1)
    else:
        size = [1] * input.ndim
    noise = torch.empty(size, dtype=input.dtype, device=input.device)
    noise = noise.bernoulli_(survival_rate)
    if survival_rate > 0.0:
        noise.div_(survival_rate)
    return input * noise


torch.fx.wrap("stochastic_depth")


class StochasticDepth(nn.Module):
    """
    See :func:`stochastic_depth`.
    """

    def __init__(self, p: float, mode: str) -> None:
        super().__init__()
        _log_api_usage_once(self)
        self.p = p
        self.mode = mode

    def forward(self, input: Tensor) -> Tensor:
        return stochastic_depth(input, self.p, self.mode, self.training)

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}(p={self.p}, mode={self.mode})"
        return s


"""Reverse-complement equivariant modules.

Taken from Caduceus repository: https://github.com/kuleshov-group/caduceus/blob/main/caduceus/modeling_rcps.py
"""

from collections import OrderedDict
from typing import Optional

import torch
from torch import Tensor
from torch import nn
from mamba_ssm.ops.triton.layer_norm import (
    RMSNorm,
    layer_norm_fn,
    rms_norm_fn,
)  # mambav2 file structure


class RCPSEmbedding(nn.Module):
    """Embedding layer that supports reverse-complement equivariance."""

    def __init__(
        self, vocab_size: int, d_model: int, complement_map: dict, **factory_kwargs
    ):
        """
        Args:
            vocab_size: Size of vocabulary.
            d_model: Dimensionality of embedding (actual embedding matrix will have 1/2 the output dim).
            complement_map: Dictionary mapping each token id to its complement.
        """
        super().__init__()
        self.register_buffer(
            "complement_map",
            torch.tensor(list(OrderedDict(complement_map).values()), dtype=torch.long),
        )
        self.embedding = nn.Embedding(vocab_size, d_model, **factory_kwargs)

    @property
    def weight(self):
        """Embedding weights."""
        return self.embedding.weight

    def set_weight(self, value):
        """Set embedding weights."""
        self.embedding.weight = value

    def rc(self, x):
        """Reverse-complement a tensor of input_ids by flipping along length dimension and complementing the ids."""
        return torch.gather(
            self.complement_map.unsqueeze(0).expand(x.shape[0], -1),
            dim=1,
            index=torch.flip(x, dims=[-1]),
        )

    def forward(self, input_ids):
        """Reverse-complement equivariant forward pass.

        This embedding module doubles the output dimensionality to support reverse-complement equivariance.

        Args:
            input_ids: Input tensor of shape (batch_size, seq_len)
        Returns:
            Embedding tensor of shape (batch_size, seq_len, d_model * 2)
        """
        fwd_out = self.embedding(input_ids)
        rc_out = torch.flip(self.embedding(self.rc(input_ids)), dims=[-2, -1])

        return torch.cat([fwd_out, rc_out], dim=-1)


class RCPSWrapper(nn.Module):
    """Wrapper to convert arbitrary nn.Module into a reverse-complement equivariant module.

    See ref. "Towards a Better Understanding of Reverse-Complement Equivariance for Deep Learning Models in Regulatory
    Genomics", Zhou et al. (2022), https://proceedings.mlr.press/v165/zhou22a.html for more details.
    """

    def __init__(self, submodule: nn.Module):
        super().__init__()
        self.submodule = submodule

    @staticmethod
    def rc(x):
        """Reverse-complement a tensor by flipping the length (dim=-2) and channel (dim=-1) dimensions."""
        return torch.flip(x, dims=[-2, -1])

    def forward(self, x, **kwargs):
        """Reverse-complement equivariant forward pass.

        Args:
            x: Input tensor of shape (batch_size, seq_len, channels)
        Returns:
            Output tensor of shape (batch_size, seq_len, channels * 2)
        """
        n_channels = x.shape[-1]
        # Run submodule along sequence
        fwd_out = self.submodule(x[..., : n_channels // 2], **kwargs)
        # Run submodule along rc-sequence
        rc_out = self.submodule(self.rc(x[..., n_channels // 2 :]), **kwargs)
        # Concatenate along channel dimension (dim=-1)
        return torch.cat([fwd_out, self.rc(rc_out)], dim=-1)


class RCPSAddNormWrapper(RCPSWrapper):
    """RC equivariant AddNorm layer."""

    def __init__(self, submodule: nn.Module):
        super().__init__(submodule)

    def forward(self, x, residual=None, prenorm=False):
        """
        Args:
            x: Input tensor of shape (batch_size, seq_len, channels)
            residual: Residual tensor of shape (batch_size, seq_len, channels) or None.
            prenorm: Whether to return residual.
        """
        n_channels = x.shape[-1]
        if residual is None:
            residual = x
            x_fwd = self.submodule(
                x[..., : n_channels // 2].to(dtype=self.submodule.weight.dtype)
            )
            x_rc = self.submodule(
                self.rc(x[..., n_channels // 2 :]).to(dtype=self.submodule.weight.dtype)
            )
            x = torch.cat([x_fwd, self.rc(x_rc)], dim=-1)
        else:
            residual_fwd = x[..., : n_channels // 2] + residual[..., : n_channels // 2]
            x_fwd = self.submodule(residual_fwd.to(dtype=self.submodule.weight.dtype))

            residual_rc = self.rc(x[..., n_channels // 2 :]) + self.rc(
                residual[..., n_channels // 2 :]
            )
            x_rc = self.submodule(residual_rc.to(dtype=self.submodule.weight.dtype))

            residual = torch.cat([residual_fwd, self.rc(residual_rc)], dim=-1)
            x = torch.cat([x_fwd, self.rc(x_rc)], dim=-1)

        return x if not prenorm else (x, residual)


class RCPSMambaBlock(nn.Module):
    def __init__(
        self,
        dim,
        mixer_cls,
        norm_cls=nn.LayerNorm,
        fused_add_norm=False,
        residual_in_fp32=False,
        device=None,  # Keep for consistency with original Mamba Block
        dtype=None,  # Keep for consistency with original Mamba Block
    ):
        """RCPS version of simple block wrapping a mixer class with LayerNorm/RMSNorm and residual connection.

        Adapted from: https://github.com/state-spaces/mamba/blob/main/mamba_ssm/modules/mamba_simple.py
        """
        super().__init__()
        self.residual_in_fp32 = residual_in_fp32
        self.fused_add_norm = fused_add_norm
        self.mixer = RCPSWrapper(mixer_cls(dim))
        norm_f = norm_cls(dim)
        self.norm = norm_f if fused_add_norm else RCPSAddNormWrapper(norm_f)
        if self.fused_add_norm:
            assert RMSNorm is not None, "RMSNorm import fails"
            assert isinstance(
                self.norm, (nn.LayerNorm, RMSNorm)
            ), "Only LayerNorm and RMSNorm are supported for fused_add_norm"

    def forward(
        self,
        hidden_states: Tensor,
        residual: Optional[Tensor] = None,
        inference_params=None,
    ):
        r"""Pass the input through the encoder layer.

        Args:
            hidden_states: the sequence to the encoder layer (required).
            residual: hidden_states = Mixer(LN(residual)).
            inference_params: inference parameters for mixer.
        """
        if not self.fused_add_norm:
            hidden_states, residual = self.norm(
                hidden_states, residual=residual, prenorm=True
            )
            if self.residual_in_fp32:
                residual = residual.to(torch.float32)
        else:
            fused_add_norm_fn = (
                rms_norm_fn if isinstance(self.norm, RMSNorm) else layer_norm_fn
            )

            hidden_states_fwd, residual_fwd = fused_add_norm_fn(
                hidden_states[..., hidden_states.shape[-1] // 2 :],
                self.norm.weight,
                self.norm.bias,
                residual=(
                    residual[..., hidden_states.shape[-1] // 2 :]
                    if residual is not None
                    else None
                ),
                prenorm=True,
                residual_in_fp32=self.residual_in_fp32,
                eps=self.norm.eps,
            )

            hidden_states_rc, residual_rc = fused_add_norm_fn(
                hidden_states[..., : hidden_states.shape[-1] // 2].flip(dims=[-2, -1]),
                self.norm.weight,
                self.norm.bias,
                residual=(
                    residual[..., : hidden_states.shape[-1] // 2].flip(dims=[-2, -1])
                    if residual is not None
                    else None
                ),
                prenorm=True,
                residual_in_fp32=self.residual_in_fp32,
                eps=self.norm.eps,
            )
            hidden_states = torch.cat(
                [hidden_states_fwd, hidden_states_rc.flip(dims=[-2, -1])], dim=-1
            )
            residual = torch.cat(
                [residual_fwd, residual_rc.flip(dims=[-2, -1])], dim=-1
            )
        hidden_states = self.mixer(hidden_states, inference_params=inference_params)
        return hidden_states, residual

    def allocate_inference_cache(self, batch_size, max_seqlen, dtype=None, **kwargs):
        """Allocate inference cache for mixer.

        Keep for compatibility with original Mamba Block.
        """
        return self.mixer.allocate_inference_cache(
            batch_size, max_seqlen, dtype=dtype, **kwargs
        )



'''

This package implements the Cell2Sen model from the open-source weights available on huggingface and the Gemma 2-2b/27b Model.
We add custom code that follows the original method outlined in the C2S-Scale paper. 
The model supports pre-processing and generating embeddings/perturbations.
See the tutorial notebook for usage. 

'''

import torch
import anndata
import numpy as np
from tqdm import tqdm
from helical.models.base_models import HelicalBaseFoundationModel
# from helical.utils.downloader import Downloader
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig
from sklearn.linear_model import LinearRegression
from datasets import Dataset
from .config import Cell2SenConfig, PERTURBATION_PROMPT, EMBEDDING_PROMPT
import logging
import torch._dynamo

LOGGER = logging.getLogger(__name__)

class Cell2Sen(HelicalBaseFoundationModel):
    """
    Cell2Sen Model.

    The Cell2Sen Model is a transformer/Gemma-based model that can be used to generate cell sentences from gene expression data.

    Example
    -------
    ```python
    from helical.models.cell2sen import Cell2Sen, Cell2SenConfig
    import anndata as ad

    config = Cell2SenConfig(batch_size=16)
    cell2sen = Cell2Sen(configurer=config)

    # Process your data
    dataloader = cell2sen.process_data(adata)

    # Get embeddings
    embeddings = cell2sen.get_embeddings(dataloader)
    print("State embeddings shape:", embeddings.shape)
    ```

    Parameters
    ----------
    configurer : Cell2SenConfig, optional, default=None
        The model configuration. If None, uses default Cell2SenConfig.

    """

    def __init__(self, configurer: Cell2SenConfig = None) -> None:
        super().__init__()

        if configurer is None:
            self.config = Cell2SenConfig().config
        else:
            self.config = configurer.config

        # downloader = Downloader()
        # for file in self.config["list_of_files_to_download"]:
        #     downloader.download_via_name(file)

        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        if self.config["dtype"] == "bfloat16":
            self.torch_dtype = torch.bfloat16
        elif self.config["dtype"] == "float32":
            self.torch_dtype = torch.float32
        else:
            raise ValueError(f"Dtype {self.config['dtype']} not supported. Please choose from 'bfloat16' or 'float32'.")

        if self.torch_dtype == torch.bfloat16 and self.device == "cpu":
            LOGGER.warning("Bfloat16 is not supported on CPU. Defaulting to 'float32' instead.")
            self.torch_dtype = torch.float32
    
        if self.config["use_quantization"]:
            self.bnb_config = BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_use_double_quant=True,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_compute_dtype=self.torch_dtype
            )
        else:
            self.bnb_config = None
       
        self.model = AutoModelForCausalLM.from_pretrained(
            self.config["hf_model_path"],
            torch_dtype=self.torch_dtype, 
            cache_dir=self.config["model_path"],
            quantization_config=self.bnb_config,
            ).to(self.device)
        
        self.tokenizer = AutoTokenizer.from_pretrained(self.config["hf_model_path"], cache_dir=self.config["model_path"])
        self.model.eval()
         
        self.batch_size = self.config['batch_size']
        self.max_new_tokens = self.config['max_new_tokens']
        self.organism = self.config['organism']
        self.perturbation_column = self.config['perturbation_column']
        self.return_fit = self.config['return_fit']
        
        LOGGER.info("Successfully loaded model")

    def process_data(
        self, 
        anndata: anndata.AnnData, 
    ):
        """
        Process anndata to create a HuggingFace Dataset with cell sentences and fit parameters.
        
        Parameters:
        -----------
        anndata : AnnData
            Annotated data object with gene expression
        min_genes : int
            Minimum number of genes expressed per cell
        min_counts : int
            Minimum total counts per cell
        min_cells : int
            Minimum number of cells expressing a gene
        max_cells : int, optional
            Maximum number of cells to process
        max_genes : int, optional
            Maximum number of genes to process
        organism : str, optional
            Organism name. If None, tries to extract from anndata.uns or uses default
        perturbation_column : str, optional
            Column name in anndata.obs to use for perturbations. If None, no perturbations are created.
        Returns:
        --------
        dataset : Dataset
            HuggingFace Dataset with fields: cell_sentence, fit_parameters, organism, perturbations
        """

        LOGGER.info("Processing data")

        X = anndata.X    
        if hasattr(X, 'toarray'):
            X = X.toarray()

        X_log = np.log10(X + 1)
        # gene names corresponding to each cell in order
        # anndata.X[i, j] is the expression of the j-th gene in the i-th cell
        gene_names = anndata.var_names.values
        cell_sentences = []
        fit_parameters = []  # Will be list of lists: [slope, intercept, r_squared] for each cell

        if self.organism is None:
            if 'organism' in anndata.uns:
                self.organism = anndata.uns['organism']
            elif 'organism' in anndata.obs.columns:
                # If organism varies per cell, use first one or most common
                self.organism = anndata.obs['organism'].iloc[0] if len(anndata.obs['organism'].unique()) == 1 else anndata.obs['organism'].mode()[0]
            elif 'species' in anndata.uns:
                self.organism = anndata.uns['species']
            elif 'species' in anndata.obs.columns:
                self.organism = anndata.obs['species'].iloc[0] if len(anndata.obs['species'].unique()) == 1 else anndata.obs['species'].mode()[0]
            else:
                self.organism = "unknown"  # Default if not found

        # Process each cell
        progress_bar = tqdm(total=X_log.shape[0], desc="Processing cells")
        for cell_idx in range(X_log.shape[0]):
            cell_expr = X_log[cell_idx, :]
            
            # Rank genes by expression (highest = rank 1)
            ranked_indices = np.argsort(cell_expr)[::-1]
            assert len(ranked_indices) != 0, "No genes expressed in cell"
            expr_values = cell_expr[ranked_indices]  # Expression values in descending order

            if self.return_fit:
                non_zero_mask = expr_values > 0
                if non_zero_mask.sum() > 0:
                    last_non_zero_idx = np.where(non_zero_mask)[0][-1]  # Last index where expr > 0
                    # Fit only up to the last non-zero gene
                    ranks_to_fit = np.arange(1, last_non_zero_idx + 2)  # +1 because rank starts at 1, +1 for inclusive
                    expr_to_fit = expr_values[:last_non_zero_idx + 1]                    
                    # Fit linear model
                    model = LinearRegression()
                    model.fit(ranks_to_fit.reshape(-1, 1), expr_to_fit)
                    slope, intercept = model.coef_[0], model.intercept_
                    r_squared = model.score(ranks_to_fit.reshape(-1, 1), expr_to_fit)
                else:
                    slope, intercept, r_squared = 0.0, 0.0, 0.0
                fit_parameters.append({"slope": float(slope), "intercept": float(intercept), "r_squared": float(r_squared)})
            else:
                fit_parameters.append(None)

            cell_sentence = " ".join(gene_names[ranked_indices])            
            cell_sentences.append(cell_sentence)
            progress_bar.update(1)
        progress_bar.close()

        if self.perturbation_column is not None:
            perturbations = anndata.obs[self.perturbation_column].values.tolist()
            if len(perturbations) != len(cell_sentences):
                raise ValueError(f"Number of perturbations ({len(perturbations)}) does not match number of cells ({len(cell_sentences)})")
        else:
            perturbations = [None] * len(cell_sentences)
        
        dataset = Dataset.from_dict({
            'cell_sentence': cell_sentences,
            'fit_parameters': fit_parameters,
            'organism': [self.organism] * len(cell_sentences),
            'perturbations': perturbations
        })

        LOGGER.info("Successfully processed data")
        
        return dataset

    def get_embeddings(
        self, 
        dataset: Dataset, 
        output_attentions: bool = False
        ):
        """
        Extract embeddings from cell sentences in a HuggingFace Dataset using the last hidden layer of Gemma.
        
        Parameters:
        -----------
        dataset : Dataset
            HuggingFace Dataset with 'cell_sentence' and 'organism' fields  
        
        output_attentions : bool, optional
            Whether to output the attention maps from the model. If set to True, the attention maps will be returned along with the embeddings.
            If set to False, only the embeddings will be returned. **Note**: This will increase the memory usage of the model significantly, so use it only if you need the attention maps.
        Returns:
        --------
        embeddings : np.ndarray
            Embeddings of shape (num_sentences, hidden_size)
        """

        LOGGER.info("Extracting embeddings from dataset")

        sentences_list = dataset['cell_sentence']
        organisms_list = dataset['organism']
        
        all_embeddings = []
        all_attentions = []

        progress_bar = tqdm(total=len(sentences_list), desc="Processing embeddings")
        for i in range(0, len(sentences_list), self.batch_size):
            batch_sentences = sentences_list[i:i + self.batch_size]
            batch_organisms = organisms_list[i:i + self.batch_size]
            
            prompts = [
                EMBEDDING_PROMPT.format(organism=org, cell_sentence=cs)
                for org, cs in zip(batch_organisms, batch_sentences)
            ]

            inputs = self.tokenizer(
                prompts,
                return_tensors="pt",
                padding=True,
                truncation=False,
                # truncation=True,
                # max_length=max_length
            ).to(self.device)

            with torch.no_grad():
                outputs = self.model(
                    **inputs, 
                    output_hidden_states=True,
                    output_attentions=output_attentions
                )
                last_hidden = outputs.hidden_states[-1]  # Shape: (batch_size, seq_len, hidden_size)
                attention_mask = inputs['attention_mask'].float() # Shape: (batch_size, seq_len, 1)

                # Sum embeddings over sequence length, masking out padding tokens
                sum_embeddings = torch.sum(last_hidden * attention_mask.unsqueeze(-1), dim=1)
                sum_mask = torch.clamp(attention_mask.sum(dim=1, keepdim=True), min=1e-9)
                batch_embeddings = sum_embeddings / sum_mask

                if output_attentions:
                    # outputs.attentions is a tuple of tensors, one per layer
                    # Each tensor has shape (batch_size, num_heads, seq_length, seq_length)
                    # Initialize all_attentions_per_layer on first batch
                    if len(all_attentions) == 0:
                        num_layers = len(outputs.attentions)
                        all_attentions = [[] for _ in range(num_layers)]
                    
                    # Append attention maps from each layer to the corresponding list
                    for layer_idx, attn in enumerate(outputs.attentions):
                        all_attentions[layer_idx].append(attn.float().cpu().numpy())
                del outputs

            all_embeddings.append(batch_embeddings.float().cpu().numpy())
            progress_bar.update(len(batch_sentences))
        progress_bar.close()
        LOGGER.info("Successfully extracted embeddings")

        if output_attentions:
            # Concatenate attention maps per layer across batches
            # Each element in stacked_attentions has shape (total_batch_size, num_heads, seq_length, seq_length)
            stacked_attentions = tuple(
                np.concatenate(all_attentions[layer_idx], axis=0)
                for layer_idx in range(len(all_attentions))
            )
            return np.concatenate(all_embeddings, axis=0), stacked_attentions
        else:
            return np.concatenate(all_embeddings, axis=0)

    def get_perturbations(
        self, 
        dataset: Dataset, 
        perturbations_list: list[str] = None, 
        ):
        """
        Generate perturbed cell sentences using the model.
        
        Parameters:
        -----------
        dataset : Dataset
            HuggingFace Dataset with 'cell_sentence' and 'perturbations' fields
        
        perturbations_list : list[str], optional
            List of perturbations to apply to the cells. If None, uses the perturbations from the dataset.
            If provided, overrides the perturbations in the dataset. E.g. ["pert1", "pert2", "pert3", ...]

        Returns:
        --------
        perturbed_dataset : Dataset
            HuggingFace Dataset with 'cell_sentence' and 'perturbations' fields and a new column 'perturbed_cell_sentence'

        perturbed_sentences : list
            List of perturbed cell sentences (strings)
        """

        LOGGER.info("Generating perturbed cell sentences")

        sentences_list = dataset['cell_sentence']
        organisms_list = dataset['organism']
        if perturbations_list is None:
            perturbations_list = dataset['perturbations']
        else:
            if len(perturbations_list) != len(sentences_list):
                raise ValueError(f"perturbations_list length ({len(perturbations_list)}) must match dataset length ({len(sentences_list)})")
        
        # Handle None perturbations - skip those entries or use empty string
        valid_indices = [i for i, p in enumerate(perturbations_list) if p is not None]
        if len(valid_indices) == 0:
                raise ValueError("No valid perturbations found in dataset. All perturbations are None.")
                    
        valid_sentences = [sentences_list[i] for i in valid_indices]
        valid_perturbations = [perturbations_list[i] for i in valid_indices]
        valid_organisms = [organisms_list[i] for i in valid_indices]    
        all_perturbed = []
        # Process in batches
        progress_bar = tqdm(total=len(valid_sentences), desc="Processing valid perturbations")
        for i in range(0, len(valid_sentences), self.batch_size):
            batch_cells = valid_sentences[i:i + self.batch_size]
            batch_perturbs = valid_perturbations[i:i + self.batch_size]
            batch_organisms = valid_organisms[i:i + self.batch_size]

            prompts = [
                PERTURBATION_PROMPT.format(
                    organism=org,
                    perturbation=pert,  # Changed from perturbation_in_words
                    cell_sentence=cs
                )
                for org, pert, cs in zip(batch_organisms, batch_perturbs, batch_cells)
            ]
            
            inputs = self.tokenizer(
                prompts,
                return_tensors="pt",
                padding=True,
                truncation=False,
                # truncation=True,
                # max_length=max_length
            ).to(self.device)
            
            with torch.no_grad():
                if self.config["use_quantization"]:
                    # Disable torch.compile entirely for quantized models
                    # suppress_errors alone isn't sufficient - we need to disable compilation
                    original_disable = torch._dynamo.config.disable
                    original_suppress = torch._dynamo.config.suppress_errors
                    torch._dynamo.config.disable = True
                    torch._dynamo.config.suppress_errors = True
                    try:
                        outputs = self.model.generate(
                            **inputs,
                            max_new_tokens=self.max_new_tokens,
                            do_sample=False,
                            pad_token_id=self.tokenizer.pad_token_id or self.tokenizer.eos_token_id
                        )
                    finally:
                        torch._dynamo.config.disable = original_disable
                        torch._dynamo.config.suppress_errors = original_suppress
                else:
                    outputs = self.model.generate(
                        **inputs,
                        max_new_tokens=self.max_new_tokens,
                        do_sample=False,
                        pad_token_id=self.tokenizer.pad_token_id or self.tokenizer.eos_token_id
                    )
            
            input_lengths = inputs['attention_mask'].sum(dim=1)
            batch_perturbed = []
            
            for j, output in enumerate(outputs):
                # Extract only the generated tokens (skip the prompt)
                input_length = input_lengths[j].item()
                generated_tokens = output[input_length:]  # Only generated part
                decoded = self.tokenizer.decode(generated_tokens, skip_special_tokens=True)
                batch_perturbed.append(decoded.strip())
            
            all_perturbed.extend(batch_perturbed)
            progress_bar.update(len(batch_cells))
        progress_bar.close()
        # Create result list with None for entries without perturbations
        perturbed_sentences = [None] * len(sentences_list)
        for idx, perturbed in zip(valid_indices, all_perturbed):
            perturbed_sentences[idx] = perturbed
        
        dataset = dataset.add_column('perturbed_cell_sentence', perturbed_sentences)

        LOGGER.info("Successfully generated perturbed cell sentences")

        return dataset, perturbed_sentences


from helical.models.uce.uce_collator import UCECollator
from torch.utils.data import Dataset
import numpy as np
import torch
from typing import Dict


class UCEDataset(Dataset):
    def __init__(
        self,
        sorted_dataset_names,
        shapes_dict,
        model_config,
        dataset_to_protein_embeddings,
        datasets_to_chroms,
        datasets_to_starts,
        expression_counts_path,
    ) -> None:
        super(UCEDataset, self).__init__()
        # self.xs = {}
        self.num_cells = {}
        self.num_genes = {}
        self.shapes_dict = shapes_dict
        self.config = model_config
        self.collator_fn = UCECollator(model_config)

        self.total_num_cells = 0
        for name in sorted_dataset_names:
            num_cells, num_genes = self.shapes_dict[name]
            # self.xs[name] = X
            self.num_cells[name] = num_cells
            self.num_genes[name] = num_genes

            self.total_num_cells += num_cells

        self.datasets = sorted_dataset_names

        # TODO: preferably not hard-coded here
        self.dataset_to_protein_embeddings = dataset_to_protein_embeddings
        self.dataset_to_chroms = datasets_to_chroms
        self.dataset_to_starts = datasets_to_starts
        self.npzs_dir = expression_counts_path

    def __getitem__(self, idx):
        if isinstance(idx, int):
            for dataset in sorted(self.datasets):
                if idx < self.num_cells[dataset]:
                    cts = np.memmap(
                        self.npzs_dir + f"{dataset}_counts.npz",
                        dtype="int64",
                        mode="r",
                        shape=self.shapes_dict[dataset],
                    )
                    counts = cts[idx]
                    counts = torch.tensor(counts).unsqueeze(0)
                    weights = torch.log1p(counts)
                    weights_sum = torch.sum(weights)
                    # zero-valued input (e.g., counts = [0, 0, 0]) must still produce a valid distribution
                    if weights_sum==0:
                        weights = torch.ones_like(weights) / len(weights)
                    else:
                        weights = weights / weights_sum
                    batch_sentences, mask, seq_len, cell_sentences = (
                        self.sample_cell_sentences(counts, weights)
                    )
                    return batch_sentences, mask, idx, seq_len, cell_sentences
                else:
                    idx -= self.num_cells[dataset]
            raise IndexError
        else:
            raise NotImplementedError

    def __len__(self) -> int:
        return self.total_num_cells

    def get_dim(self) -> Dict[str, int]:
        return self.num_genes

    def sample_cell_sentences(self, counts, batch_weights):

        dataset_idxs = (
            self.dataset_to_protein_embeddings
        )  # get the dataset specific protein embedding idxs
        cell_sentences = torch.zeros(
            (counts.shape[0], self.config["pad_length"])
        )  # init the cell representation as 0s
        mask = torch.zeros(
            (counts.shape[0], self.config["pad_length"])
        )  # start of masking the whole sequence
        chroms = self.dataset_to_chroms  # get the dataset specific chroms for each gene
        starts = (
            self.dataset_to_starts
        )  # get the dataset specific genomic start locations for each gene

        longest_seq_len = (
            0  # we need to keep track of this so we can subset the batch at the end
        )
        for c, cell in enumerate(counts):
            weights = batch_weights[c].numpy()
            weights = weights / sum(weights)  # RE NORM after mask

            # set the seed for reproducibility
            generator = np.random.default_rng(seed=42)

            # randomly choose the genes that will make up the sample, weighted by expression, with replacement
            choice_idx = generator.choice(
                np.arange(len(weights)),
                size=self.config["sample_size"],
                p=weights,
                replace=True,
            )

            choosen_chrom = chroms.iloc[choice_idx]  # get the sampled genes chromosomes
            # order the genes by chromosome
            chrom_sort = np.argsort(choosen_chrom)
            choice_idx = choice_idx[chrom_sort]

            # sort the genes by start
            new_chrom = chroms.iloc[choice_idx]
            choosen_starts = starts[choice_idx]

            ordered_choice_idx = np.full(
                (self.config["pad_length"]), self.config["cls_token_idx"]
            )  # start with cls
            # i= 0 first token is CLS
            i = 1  # continue on to the rest of the sequence with left bracket being assumed.
            # Shuffle the chroms now, there's no natural order to chromosomes
            uq_chroms = np.unique(new_chrom)
            generator.shuffle(uq_chroms)  # shuffle

            # This loop is actually just over one cell
            for chrom in uq_chroms:
                # Open Chrom token
                ordered_choice_idx[i] = (
                    int(chrom) + self.config["CHROM_TOKEN_OFFSET"]
                )  # token of this chromosome # i = 1 next token is a chrom open
                i += 1
                # now sort the genes by start order within the chroms
                loc = np.where(new_chrom == chrom)[0]
                sort_by_start = np.argsort(
                    choosen_starts[loc]
                )  # start locations for this chromsome

                to_add = choice_idx[loc[sort_by_start]]
                ordered_choice_idx[i : (i + len(to_add))] = dataset_idxs[to_add]
                i += len(to_add)
                ordered_choice_idx[i] = self.config[
                    "chrom_token_right_idx"
                ]  # add the chrom sep again
                i += 1  # add the closing token again

            longest_seq_len = max(longest_seq_len, i)
            remainder_len = self.config["pad_length"] - i

            cell_mask = torch.concat(
                (
                    torch.ones(i),
                    # pay attention to all of these tokens, ignore the rest!
                    torch.zeros(remainder_len),
                )
            )

            mask[c, :] = cell_mask

            ordered_choice_idx[i:] = self.config[
                "pad_token_idx"
            ]  # the remainder of the sequence
            cell_sentences[c, :] = torch.from_numpy(ordered_choice_idx)

        cell_sentences_pe = cell_sentences.long()  # token indices

        return cell_sentences_pe, mask, longest_seq_len, cell_sentences



import logging
from typing import List, Optional, Sequence
from helical.utils.downloader import Downloader
import pandas as pd
from anndata import AnnData
import pybiomart
from pathlib import Path
from helical.constants.paths import CACHE_DIR_HELICAL

LOGGER = logging.getLogger(__name__)


def _get_ensembl_mart_df(species: str = "hsapiens") -> pd.DataFrame:
    """
    Fetch a (species)_gene_ensembl table via pybiomart.

    Parameters
    ----------
    species : str, default "hsapiens"
        Species prefix used by Ensembl Biomart (e.g., "hsapiens", "mmusculus").

    Returns
    -------
    pandas.DataFrame
        DataFrame with columns "ensembl_id" and "gene_name".
    """
    server = pybiomart.Server(host="http://www.ensembl.org")
    dataset = server.marts["ENSEMBL_MART_ENSEMBL"].datasets[f"{species}_gene_ensembl"]
    df = dataset.query(attributes=["ensembl_gene_id", "external_gene_name"])
    df = df.rename(columns={"Gene stable ID": "ensembl_id", "Gene name": "gene_name"})
    return df.sort_values(by="ensembl_id")


def map_gene_symbols_to_ensembl_ids(
    adata: AnnData, gene_names: Optional[str] = None, species: str = "hsapiens"
) -> AnnData:
    """
    Map gene symbols to Ensembl Gene IDs using pybiomart.

    Due to duplication events, some symbols map to multiple Ensembl IDs; we take
    the first occurrence after de-duplication.

    Parameters
    ----------
    adata : AnnData
        AnnData object containing gene metadata in `adata.var`.
    gene_names : str, optional
        Column in `adata.var` containing gene symbols. If None, uses `adata.var_names`.
    species : str, default "hsapiens"
        Species prefix used by Ensembl Biomart (e.g., "hsapiens", "mmusculus").

    Returns
    -------
    AnnData
        Same object with `adata.var["ensembl_id"]` populated.
    """
    var_names = adata.var[gene_names] if gene_names is not None else pd.Series(adata.var_names, index=adata.var_names)
    adata.var["ensembl_id"] = convert_list_gene_symbols_to_ensembl_ids(var_names, species=species)
    non_none_mappings = adata.var["ensembl_id"].notnull().sum()
    LOGGER.info("Mapped %d / %d genes to Ensembl IDs.", non_none_mappings, adata.var.shape[0])
    return adata


def map_ensembl_ids_to_gene_symbols(
    adata: AnnData, ensembl_id_key: str = "ensembl_id", species: str = "hsapiens"
) -> AnnData:
    """
    Map Ensembl Gene IDs to gene symbols using pybiomart.

    Parameters
    ----------
    adata : AnnData
        AnnData object containing gene metadata in `adata.var`.
    ensembl_id_key : str, default "ensembl_id"
        Column in `adata.var` containing Ensembl Gene IDs.
    species : str, default "hsapiens"
        Species prefix used by Ensembl Biomart (e.g., "hsapiens", "mmusculus").

    Returns
    -------
    AnnData
        Same object with `adata.var["gene_names"]` populated.
    """
    adata.var["gene_names"] = convert_list_ensembl_ids_to_gene_symbols(adata.var[ensembl_id_key], species=species)
    non_none_mappings = adata.var["gene_names"].notnull().sum()
    LOGGER.info("Mapped %d / %d Ensembl IDs to gene names.", non_none_mappings, adata.var.shape[0])
    return adata

def _load_static_ensembl_df() -> pd.DataFrame:
    """
    Load a static mapping table between gene names and ensembl ids for 'hsapiens'.
    This avoids having to call an unstable API endpoint from pybiomart.
    Instead load a static csv from helical.

    Returns
    -------
    pandas.DataFrame
        DataFrame with columns "ensembl_id" and "gene_name".
    """
    downloader = Downloader()
    downloader.download_via_name('hsapiens_pybiomart.csv')
    df_path = Path(CACHE_DIR_HELICAL, "hsapiens_pybiomart.csv")
    df = pd.read_csv(df_path, index_col=0)
    return df


def convert_list_ensembl_ids_to_gene_symbols(
    ensembl_ids: Sequence[str], species: str = "hsapiens"
) -> List[Optional[str]]:
    """
    Map a list/sequence of Ensembl Gene IDs to gene symbols using pybiomart.

    Parameters
    ----------
    ensembl_ids : Sequence[str]
        Ensembl Gene IDs (e.g., "ENSG00000139618").
    species : str, default "hsapiens"
        Species prefix used by Ensembl Biomart.

    Returns
    -------
    List[Optional[str]]
        Gene symbols aligned to the input order (None if not found).
    """
    if species == "hsapiens":
        df = _load_static_ensembl_df()
    else:
        df = _get_ensembl_mart_df(species=species)
    mapping = df.drop_duplicates(subset="ensembl_id").set_index("ensembl_id")["gene_name"]
    return list(pd.Series(list(ensembl_ids), dtype="object").map(mapping).where(pd.notna, None))


def convert_list_gene_symbols_to_ensembl_ids(
    gene_symbols: Sequence[str], species: str = "hsapiens"
) -> List[Optional[str]]:
    """
    Map a list/sequence of gene symbols to Ensembl Gene IDs using pybiomart.

    Parameters
    ----------
    gene_symbols : Sequence[str]
        Gene symbols (e.g., "BRCA2", "KANSL2").
    species : str, default "hsapiens"
        Species prefix used by Ensembl Biomart.

    Returns
    -------
    List[Optional[str]]
        Ensembl Gene IDs aligned to the input order (None if not found).
    """
    if species == "hsapiens":
        df = _load_static_ensembl_df()
    else:
        df = _get_ensembl_mart_df(species=species)
    mapping = df.drop_duplicates(subset="gene_name").set_index("gene_name")["ensembl_id"]
    return list(pd.Series(list(gene_symbols), dtype="object").map(mapping).where(pd.notna, None))


from typing import Literal
import logging

LOGGER = logging.getLogger(__name__)


class Mamba2mRNAConfig:
    """Mamba2-mRNA Config class to store the configuration of the Mamba2-mRNA model.

    Parameters
    ----------
    batch_size : int, optional, default=10
        The batch size
    device : Literal["cpu", "cuda"], optional, default="cpu"
        The device to use. Either use "cuda" or "cpu".
    max_length : int, optional, default=8192
        The maximum length of the input sequence. It is recommeded to set this to the longest sequence in the dataset.
    nproc: int, optional, default=1
        Number of processes to use for data processing.
    """

    def __init__(
        self,
        batch_size: int = 10,
        device: Literal["cpu", "cuda"] = "cpu",
        max_length: int = 8192,
        nproc: int = 1,
    ):

        model_name = "helical-ai/mamba2-mRNA"

        self.config = {
            "model_name": model_name,
            "batch_size": batch_size,
            "input_size": max_length,
            "device": device,
            "nproc": nproc,
        }



# coding=utf-8
# Copyright 2024 The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""HelixmRNA configuration"""

import math

from transformers.configuration_utils import PretrainedConfig
from transformers.utils import logging


logger = logging.get_logger(__name__)


class HelixmRNAPretrainedConfig(PretrainedConfig):
    """
    This is the configuration class to store the configuration of a [`HelixmRNAModel`]. It is used to instantiate a Helix-mRNA
    model according to the specified arguments, defining the model architecture.

    Configuration objects inherit from [`PretrainedConfig`] and can be used to control the model outputs. Read the
    documentation from [`PretrainedConfig`] for more information.


    Args:
        num_heads (`int`, *optional*, defaults to 128):
            Number of heads for the evolution matrices of mamba 2.
        head_dim (`int`, *optional*, defaults to 64):
            Dimension of each head.
        vocab_size (`int`, *optional*, defaults to 32768):
            Vocabulary size of the Helix-mRNA model. Defines the number of different tokens that can be represented by the
            `inputs_ids` passed when calling [`HelixmRNAModel`].
        hidden_size (`int`, *optional*, defaults to 4096):
            Dimensionality of the embeddings and hidden states.
        state_size (`int`, *optional*, defaults to 128):
            Shape of the state space latents.
        num_hidden_layers (`int`, *optional*, defaults to 64):
            Number of hidden layers in the model.
        num_attention_heads (`int`, *optional*, defaults to 32):
            Number of attention heads for the attention layers.
        num_key_value_heads (`int`, *optional*, defaults to 8):
            Number of key and value heads for the attention layers.
        layer_norm_epsilon (`float`, *optional*, defaults to 1e-05):
            The epsilon to use in the layer normalization layers.
        pad_token_id (`int`, *optional*, defaults to 1):
            Padding token id.
        bos_token_id (`int`, *optional*, defaults to 0):
            The id of the beginning of sentence token in the vocabulary.
        eos_token_id (`int`, *optional*, defaults to 2):
            The id of the end of sentence token in the vocabulary.
        expand (`int`, *optional*, defaults to 2):
            Expanding factor used to determine the intermediate size.
        conv_kernel (`int`, *optional*, defaults to 4):
            Size of the convolution kernel.
        n_groups (`int`, *optional*, defaults to 8):
            Number of groups for the evolution matrices of the Mamba2 block.
        use_bias (`bool`, *optional*, defaults to `False`):
            Whether or not to use bias in ["in_proj", "out_proj"] of the mixer block
        use_conv_bias (`bool`, *optional*, defaults to `True`):
            Whether or not to use bias in the convolution layer of the mixer block.
        hidden_act (`str`, *optional*, defaults to `"silu"`):
            The non-linear activation function (function or string) in the decoder.
        initializer_range (`float`, *optional*, defaults to 0.1):
            The standard deviation of the truncated_normal_initializer for initializing all weight matrices.
        residual_in_fp32 (`bool`, *optional*, defaults to `True`):
            Whether or not residuals should be in `float32`. If set to `False` residuals will keep the same `dtype` as the rest of the model
        time_step_rank (`Union[int,str]`, *optional*, defaults to `"auto"`):
            Rank of the discretization projection matrix. `"auto"` means that it will default to `math.ceil(self.hidden_size / 16)`
        time_step_min (`float`, *optional*, defaults to 0.001):
            Minimum `time_step` used to bound `dt_proj.bias`.
        time_step_max (`float`, *optional*, defaults to 0.1):
            Maximum `time_step` used to bound `dt_proj.bias`.
        time_step_floor (`float`, *optional*, defaults to 0.0001):
            Minimum clamping value of the `dt_proj.bias` layer initialization.
        time_step_limit (`tuple`, *optional*, defaults to `(0.0, inf)`):
            Accepted range of time step values.
        rescale_prenorm_residual (`bool`, *optional*, defaults to `False`):
            Whether or not to rescale `out_proj` weights when initializing.
        use_cache (`bool`, *optional*, defaults to `True`):
            Whether or not the cache should be used.
        rms_norm (`bool`, *optional*, defaults to `True`):
            Whether to use RMS norm or not.
        chunk_size (`int`, *optional*, defaults to 256):
            Size of the chunks that will comprise the sequence.
        tie_word_embeddings (`bool`, *optional*, defaults to `False`):
            Whether to tie word embeddings or not.
        attention_dropout (`float`, *optional*, defaults to 0.0):
            The dropout probability for the attention layers.
        layers_block_type (`str`, *optional*, defaults to `"M+M+M+M*+M+M+M+M+M+M+M+M*+M+M+M+M"`):
            The type of layers in the model. The string should contain the following characters:
            - `"M"`: Mamba2Block
            - `"*"`: HelixmRNASdpaAttention
            - `"+"`: HelixmRNAMLP
        intermediate_size (`int`, *optional*, defaults to 512):
            Intermediate size of the MLP layer.
    """

    def __init__(
        self,
        num_heads=128,
        head_dim=64,
        vocab_size=32768,
        hidden_size=4096,
        state_size=128,
        num_hidden_layers=64,
        num_attention_heads=32,
        num_key_value_heads=8,
        layer_norm_epsilon=1e-5,
        pad_token_id=1,
        bos_token_id=0,
        eos_token_id=2,
        expand=2,
        conv_kernel=4,
        n_groups=8,
        use_bias=False,
        use_conv_bias=True,
        hidden_act="silu",
        initializer_range=0.1,
        residual_in_fp32=True,
        time_step_rank="auto",
        time_step_min=0.001,
        time_step_max=0.1,
        time_step_floor=1e-4,
        time_step_limit=(0.0, float("inf")),
        rescale_prenorm_residual=False,
        use_cache=True,
        rms_norm=True,
        chunk_size=256,
        tie_word_embeddings=False,
        attention_dropout=0.0,
        layers_block_type="M+M+M+M*+M+M+M+M+M+M+M+M*+M+M+M+M",
        intermediate_size=512,
        **kwargs,
    ):
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.state_size = state_size
        self.num_hidden_layers = num_hidden_layers
        self.layer_norm_epsilon = layer_norm_epsilon
        self.conv_kernel = conv_kernel
        self.expand = expand

        self.num_attention_heads = num_attention_heads
        # for backward compatibility
        if num_key_value_heads is None:
            num_key_value_heads = num_attention_heads
        self.num_key_value_heads = num_key_value_heads
        self.attention_dropout = attention_dropout

        self.bos_token_id = bos_token_id
        self.eos_token_id = eos_token_id
        self.pad_token_id = pad_token_id
        self.use_bias = use_bias
        self.use_conv_bias = use_conv_bias
        self.hidden_act = hidden_act
        self.initializer_range = initializer_range
        self.time_step_rank = (
            math.ceil(self.hidden_size / 16)
            if time_step_rank == "auto"
            else time_step_rank
        )
        self.time_step_min = time_step_min
        self.time_step_max = time_step_max
        self.time_step_floor = time_step_floor
        self.rescale_prenorm_residual = rescale_prenorm_residual
        self.residual_in_fp32 = residual_in_fp32
        self.use_cache = use_cache
        self.n_groups = n_groups
        self.num_heads = num_heads
        self.head_dim = head_dim
        self.rms_norm = rms_norm
        self.state_size = state_size
        self.chunk_size = chunk_size
        self.time_step_limit = time_step_limit
        self.tie_word_embeddings = tie_word_embeddings
        self.layers_block_type_string = layers_block_type
        self.intermediate_size = intermediate_size

        super().__init__(
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            tie_word_embeddings=tie_word_embeddings,
            **kwargs,
        )

    @property
    def layers_block_type(self):
        layers = []
        if self.num_hidden_layers != len(self.layers_block_type_string):
            raise ValueError(
                f"num_hidden_layers should be equal to the number of layers in layers_block_type_string, but got {self.num_hidden_layers} and {len(self.layers_block_type_string)}"
            )
        for layer in self.layers_block_type_string:
            if layer == "M":
                layers.append("mamba")
            elif layer == "*":
                layers.append("attention")
            elif layer == "+":
                layers.append("mlp")
        return layers



from helical.models.geneformer import Geneformer, GeneformerConfig
from helical.models.scgpt import scGPT, scGPTConfig
import hydra
from omegaconf import DictConfig
import anndata as ad
from datasets import load_dataset
from helical.utils import get_anndata_from_hf_dataset
import time
import numpy as np
import torch


def run_geneformer(cfg: DictConfig, ann_data: ad.AnnData):
    geneformer_config = GeneformerConfig(**cfg)
    geneformer = Geneformer(configurer=geneformer_config)

    num = geneformer.model.num_parameters(only_trainable=False)
    print(f"Number of parameters: {num:_}".replace("_", " "))

    # either load via huggingface
    # hf_dataset = load_dataset(
    #     "helical-ai/yolksac_human",
    #     split="train[:5%]",
    #     trust_remote_code=True,
    #     download_mode="reuse_cache_if_exists",
    # )
    # ann_data = get_anndata_from_hf_dataset(hf_dataset)

    # or load directly

    start_time = time.time()
    dataset = geneformer.process_data(ann_data[:])
    end_time = time.time()
    print(f"Data processing time: {end_time - start_time:.4f} seconds")
    start_time = time.time()
    embeddings, input_genes = geneformer.get_embeddings(dataset, output_genes=True)
    end_time = time.time()
    print(
        f"Get embeddings time (output_genes=True): {end_time - start_time:.4f} seconds"
    )
    # np.save("geneformer_embeddings.npy", embeddings)
    print(embeddings)
    # print(embeddings)
    # embeddings, attention_weights = geneformer.get_embeddings(
    #     dataset, output_attentions=True
    # )


def run_scgpt(cfg: DictConfig, ann_data: ad.AnnData):
    scgpt_config = scGPTConfig(**cfg)
    scgpt = scGPT(configurer=scgpt_config)

    # either load via huggingface
    # hf_dataset = load_dataset(
    #     "helical-ai/yolksac_human",
    #     split="train[:5%]",
    #     trust_remote_code=True,
    #     download_mode="reuse_cache_if_exists",
    # )
    # ann_data = get_anndata_from_hf_dataset(hf_dataset)

    # or load directly
    # ann_data = ad.read_h5ad("./yolksac_human.h5ad")
    start_time = time.time()
    data = scgpt.process_data(ann_data[:])
    end_time = time.time()
    print(f"Data processing time: {end_time - start_time:.4f} seconds")
    start_time = time.time()
    embeddings, input_genes = scgpt.get_embeddings(data, output_genes=True)

    end_time = time.time()
    print(
        f"Get embeddings time (output_genes=True): {end_time - start_time:.4f} seconds"
    )
    # np.save("scgpt_embeddings.npy", embeddings)
    print(embeddings)
    # embeddings, attn_weights = scgpt.get_embeddings(data, output_attentions=True)


if __name__ == "__main__":

    ## Load Specific Hydra Config
    ann_data = ad.read_h5ad("./yolksac_human.h5ad")
    hydra.initialize(config_path="configs", job_name="geneformer_config")
    cfg_geneformer = hydra.compose(
        config_name="geneformer_config", overrides=["device=cuda"]
    )
    start_time = time.time()
    run_geneformer(cfg_geneformer, ann_data)
    end_time = time.time()
    print(f"Geneformer run time: {end_time - start_time:.4f} seconds")
    print("\n\n\n")

    torch.cuda.empty_cache()
    # hydra.initialize(config_path="configs", job_name="scgpt_config")
    cfg_scgpt = hydra.compose(config_name="scgpt_config", overrides=["device=cuda"])
    start_time = time.time()
    run_scgpt(cfg_scgpt, ann_data)
    end_time = time.time()
    print(f"scGPT run time: {end_time - start_time:.4f} seconds")



"""Caduceus config for Hugging Face.

Taken from Caduceus repository: https://github.com/kuleshov-group/caduceus/blob/main/caduceus/configuration_caduceus.py
"""

from typing import Optional, Union

from transformers import PretrainedConfig


class CaduceusPretrainedConfig(PretrainedConfig):
    """Config that extends the original MambaConfig with params relevant to bi-directionality and RC equivariance.

    This config is used to initialise the Caduceus pretrained model.

    Parameters
    ----------
    d_model : int, optional, default=2560
        The model dimension.
    n_layer : int, optional, default=64
        The number of layers.
    vocab_size : int, optional, default=50277
        The vocabulary size.
    ssm_cfg : Optional[dict], optional, default=None
        The configuration for the SSM.
    rms_norm : bool, optional, default=True
        Whether to use RMS normalization.
    residual_in_fp32 : bool, optional, default=True
        Whether to use residual connections in FP32.
    fused_add_norm : bool, optional, default=True
        Whether to use fused add norm.
    pad_vocab_size_multiple : int, optional, default=8
        The padding vocabulary size multiple.
    norm_epsilon : float, optional, default=1e-5
        The epsilon value for layer normalization.
    initializer_cfg : Optional[dict], optional, default=None
        The configuration for the initializer.
    bidirectional : bool, optional, default=True
        Whether to use bidirectional attention.
    bidirectional_strategy : Union[str, None], optional, default="add"
        The bidirectional strategy to use.
    bidirectional_weight_tie : bool, optional, default=True
        Whether to tie the weights for bidirectional attention.
    rcps : bool, optional, default=False
        Whether to use RCPS.
    complement_map : Optional[dict], optional, default=None
        The complement map.
    """

    def __init__(
        self,
        # From original MambaConfig
        d_model: int = 2560,
        n_layer: int = 64,
        vocab_size: int = 50277,
        ssm_cfg: Optional[dict] = None,
        rms_norm: bool = True,
        residual_in_fp32: bool = True,
        fused_add_norm: bool = True,
        pad_vocab_size_multiple: int = 8,
        # Not in original MambaConfig, but default arg in create_block in mamba_ssm repo; used in layer norm
        norm_epsilon: float = 1e-5,
        # Used in init_weights
        initializer_cfg: Optional[dict] = None,
        # Caduceus-specific params
        bidirectional: bool = True,
        bidirectional_strategy: Union[str, None] = "add",
        bidirectional_weight_tie: bool = True,
        rcps: bool = False,
        complement_map: Optional[dict] = None,  # used for RCPSEmbedding / RCPSLMHead
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.d_model = d_model
        self.n_layer = n_layer
        self.vocab_size = vocab_size
        self.ssm_cfg = ssm_cfg
        self.rms_norm = rms_norm
        self.residual_in_fp32 = residual_in_fp32
        self.fused_add_norm = fused_add_norm
        self.pad_vocab_size_multiple = pad_vocab_size_multiple
        self.norm_epsilon = norm_epsilon
        self.initializer_cfg = initializer_cfg
        self.bidirectional = bidirectional
        self.bidirectional_strategy = bidirectional_strategy
        self.bidirectional_weight_tie = bidirectional_weight_tie
        self.rcps = rcps
        self.complement_map = complement_map



from helical.models.evo_2 import Evo2, Evo2Config
import hydra
from omegaconf import DictConfig
from pandas import DataFrame


@hydra.main(version_base=None, config_path="configs", config_name="evo_2_config")
def run(cfg: DictConfig):
    # Load the Evo2 model
    evo2_config = Evo2Config(**cfg)
    evo2 = Evo2(configurer=evo2_config)

    data = DataFrame({"Sequence": ["ACTG" * int(1024 / 4), "TGCA" * int(1024 / 2)]})

    dataset = evo2.process_data(data)

    embeddings = evo2.get_embeddings(dataset)

    generate = evo2.generate(dataset)
    # Get the last embedding of each sequence
    print(embeddings["embeddings"][0][embeddings["original_lengths"][0]-1])
    print(embeddings["embeddings"][1][embeddings["original_lengths"][1]-1])
    print(embeddings["original_lengths"])

    # Print the generated sequences
    print(generate)

if __name__ == "__main__":
    run()


"""
Model class

"""

import warnings

warnings.filterwarnings("ignore")
import math
from torch import nn, Tensor
from torch.nn import TransformerEncoder, TransformerEncoderLayer

import sys

sys.path.append("../")
from typing import Any
import torch


def full_block(in_features, out_features, p_drop=0.1):
    return nn.Sequential(
        nn.Linear(in_features, out_features, bias=True),
        nn.LayerNorm(out_features),
        nn.GELU(),
        nn.Dropout(p=p_drop),
    )


class PositionalEncoding(nn.Module):

    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 1536):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model)
        )
        pe = torch.zeros(max_len, 1, d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe)

    def forward(self, x: Tensor) -> Tensor:
        """
        Args:
            x: Tensor, shape [seq_len, batch_size, embedding_dim]
        """
        x = x + self.pe[: x.size(0)]
        return self.dropout(x)


class TransformerModel(nn.Module):

    def __init__(
        self,
        token_dim: int,
        d_model: int,
        nhead: int,
        d_hid: int,
        nlayers: int,
        output_dim: int,
        dropout: float = 0.05,
    ):
        super().__init__()
        self.model_type = "Transformer"
        self.pos_encoder = PositionalEncoding(d_model, dropout)
        self.d_model = d_model

        self.encoder = nn.Sequential(
            nn.Linear(token_dim, d_model), nn.GELU(), nn.LayerNorm(d_model)
        )

        encoder_layers = TransformerEncoderLayer(d_model, nhead, d_hid, dropout)
        self.transformer_encoder = TransformerEncoder(encoder_layers, nlayers)

        self.d_model = d_model
        self.dropout = dropout

        self.decoder = nn.Sequential(
            full_block(d_model, 1024, self.dropout),
            full_block(1024, output_dim, self.dropout),
            full_block(output_dim, output_dim, self.dropout),
            nn.Linear(output_dim, output_dim),
        )

        self.binary_decoder = nn.Sequential(
            full_block(output_dim + 1280, 2048, self.dropout),
            full_block(2048, 512, self.dropout),
            full_block(512, 128, self.dropout),
            nn.Linear(128, 1),
        )

        self.gene_embedding_layer = nn.Sequential(
            nn.Linear(token_dim, d_model), nn.GELU(), nn.LayerNorm(d_model)
        )

        self.pe_embedding = None

    def forward(self, src: Tensor, mask: Tensor):
        """
        Args:
            src: Tensor, shape [seq_len, batch_size]
        Returns:
            output Tensor of shape [seq_len, batch_size, ntoken]
        """
        src = self.encoder(src) * math.sqrt(self.d_model)
        src = self.pos_encoder(src)
        output = self.transformer_encoder(src, src_key_padding_mask=(1 - mask))
        gene_output = self.decoder(output)  # batch x seq_len x 128
        # embedding = torch.mul(gene_output, mask.t().unsqueeze(2)).sum(0) # average over non zero genes
        # In the new format, the cls token, which is at the 0 index mark, is the output.
        embedding = gene_output[0, :, :]  # select only the CLS token.
        embedding = nn.functional.normalize(embedding, dim=1)  # Normalize.
        return gene_output, embedding

    def predict(self, cell_embedding, gene_embeddings):
        gene_embeddings = self.gene_embedding_layer(gene_embeddings)
        dec = self.binary_decoder(torch.hstack((cell_embedding, gene_embeddings)))
        return dec



from helical.models.uce.uce_dataset import UCEDataset
import numpy as np
import torch
from torch import optim
from torch.nn.modules import loss
from torch.utils.data import DataLoader
from helical.models.base_models import HelicalBaseFineTuningHead
from helical.models.base_models import HelicalBaseFineTuningModel
from helical.models.uce import UCE, UCEConfig
from typing import Literal, Optional
from tqdm import tqdm
from transformers import get_scheduler
import logging

logger = logging.getLogger(__name__)


class UCEFineTuningModel(HelicalBaseFineTuningModel, UCE):
    """
    Fine-tuning model for the UCE model.

    Example
    ----------
    ```python
    from helical.models.uce import UCEConfig, UCEFineTuningModel
    import anndata as ad

    # Load the data
    ann_data = ad.read_h5ad("dataset.h5ad")

    # Get unique output labels
    label_set = set(cell_types)

    # Create the fine-tuning model with the desired configs
    configurer=UCEConfig(batch_size=10)
    uce_fine_tune = UCEFineTuningModel(uce_config=configurer, fine_tuning_head="classification", output_size=len(label_set))

    # Process the data for training
    dataset = uce_fine_tune.process_data(ann_data)

    # Get the desired label class
    cell_types = list(ann_data.obs.cell_type)

    # Create a dictionary mapping the classes to unique integers for training
    class_id_dict = dict(zip(label_set, [i for i in range(len(label_set))]))

    for i in range(len(cell_types)):
        cell_types[i] = class_id_dict[cell_types[i]]

    # Fine-tune
    uce_fine_tune.train(train_input_data=dataset, train_labels=cell_types)
    ```

    Parameters
    ----------
    uce_config : UCE
        The UCE configs for fine-tuning model, the same configs that would be used to instantiate the standard UCE model.
    fine_tuning_head : Literal["classification", "regression"] | HelicalBaseFineTuningHead
        The fine-tuning head that is appended to the model. This can either be a string (options available: "classification", "regression") specifying the task or a custom fine-tuning head inheriting from HelicalBaseFineTuningHead.
    output_size : Optional[int]
        The output size of the fine-tuning model. This is required if the fine_tuning_head is a string specified task. For a classification task this is number of unique classes.

    Methods
    -------
    train(train_input_data: UCEDataset, train_labels: np.ndarray, validation_input_data = None, validation_labels = None, optimizer: optim = optim.AdamW, optimizer_params: dict = {'lr': 0.0001}, loss_function: loss = loss.CrossEntropyLoss(), epochs: int = 1, lr_scheduler_params: Optional[dict] = None)
        Fine-tunes the UCE model with different head modules.
    get_outputs(dataset: UCEDataset) -> np.ndarray
        Get the outputs of the fine-tuned model on a UCE processed dataset.

    """

    def __init__(
        self,
        uce_config: UCEConfig,
        fine_tuning_head: Literal["classification"] | HelicalBaseFineTuningHead,
        output_size: Optional[int] = None,
    ):

        HelicalBaseFineTuningModel.__init__(self, fine_tuning_head, output_size)
        UCE.__init__(self, uce_config)

        self.fine_tuning_head.set_dim_size(self.config["embsize"])

    def _forward(
        self, batch_sentences: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """
        Forward method of the fine-tuning model.

        Parameters
        ----------
        batch_sentences : torch.Tensor
            The input tensor of the fine-tuning model.
        mask : torch.Tensor
            The mask tensor for the input tensor.

        Returns
        -------
        torch.Tensor
            The output tensor of the fine-tuning model.
        """
        _, embeddings = self.model.forward(batch_sentences, mask=mask)
        if self.accelerator is not None:
            self.accelerator.wait_for_everyone()
            embeddings = self.accelerator.gather_for_metrics((embeddings))
            if self.accelerator.is_main_process:
                embeddings = embeddings
        else:
            embeddings = embeddings
        output = self.fine_tuning_head(embeddings)
        return output

    def train(
        self,
        train_input_data: UCEDataset,
        train_labels: np.ndarray,
        validation_input_data=None,
        validation_labels=None,
        optimizer: optim = optim.AdamW,
        optimizer_params: dict = {"lr": 0.0001},
        loss_function: loss = loss.CrossEntropyLoss(),
        epochs: int = 1,
        # freeze_layers: int = 0,
        lr_scheduler_params: Optional[dict] = None,
    ):
        """
        Fine-tunes the UCE model with different head modules.

        Parameters
        ----------
        train_input_data : Dataset
            A helical UCE processed dataset for fine-tuning
        train_labels : ndarray
            The labels for the training data. These should be stored as unique per class integers.
        validation_input_data : Dataset, default=None
            A helical UCE processed dataset for per epoch validation. If this is not specified, no validation will be performed.
        validation_labels : ndarray, default=None,
            The labels for the validation data. These should be stored as unique per class integers.
        optimizer : torch.optim, default=torch.optim.AdamW
            The optimizer to be used for training.
        optimizer_params : dict
            The optimizer parameters to be used for the optimizer specified. This list should NOT include model parameters.
            e.g. optimizer_params={'lr': 0.0001}
        loss_function : torch.nn.modules.loss, default=torch.nn.CrossEntropyLoss()
            The loss function to be used.
        epochs : int, optional, default=10
            The number of epochs to train the model
        lr_scheduler_params : dict, default=None
            The learning rate scheduler parameters for the transformers get_scheduler method. The optimizer will be taken from the optimizer input and should not be included in the learning scheduler parameters. If not specified, no scheduler will be used.
            e.g. lr_scheduler_params={'name': 'linear', 'num_warmup_steps': 0, 'num_training_steps': 5}
        """
        batch_size = self.config["batch_size"]
        dataloader = DataLoader(
            train_input_data,
            batch_size=batch_size,
            shuffle=False,
            collate_fn=train_input_data.collator_fn,
            num_workers=0,
        )

        if validation_input_data is not None:
            validation_dataloader = DataLoader(
                validation_input_data,
                batch_size=batch_size,
                shuffle=False,
                collate_fn=validation_input_data.collator_fn,
                num_workers=0,
            )

        if self.accelerator is not None:
            dataloader = self.accelerator.prepare(dataloader)
            if validation_input_data is not None:
                validation_dataloader = self.accelerator.prepare(validation_dataloader)

        self.model.train()
        self.fine_tuning_head.train()

        # disable progress bar if not the main process
        # if self.accelerator is not None:
        #     pbar = tqdm(dataloader, disable=not self.accelerator.is_local_main_process)
        # else:
        #     pbar = tqdm(dataloader)

        self.to(self.device)

        optimizer = optimizer(self.parameters(), **optimizer_params)

        lr_scheduler = None
        if lr_scheduler_params is not None:
            lr_scheduler = get_scheduler(optimizer=optimizer, **lr_scheduler_params)

        logger.info("Starting Fine-Tuning")
        for j in range(epochs):
            batch_count = 0
            batch_loss = 0.0
            batches_processed = 0
            training_loop = tqdm(dataloader, desc="Fine-Tuning")
            for batch in training_loop:
                batch_sentences, mask, idxs = batch[0], batch[1], batch[2]
                batch_sentences = batch_sentences.permute(1, 0)
                if self.config["multi_gpu"]:
                    batch_sentences = self.model.module.pe_embedding(
                        batch_sentences.long()
                    )
                else:
                    batch_sentences = self.model.pe_embedding(batch_sentences.long())
                batch_sentences = torch.nn.functional.normalize(
                    batch_sentences, dim=2
                )  # normalize token outputs
                output = self._forward(batch_sentences, mask=mask)
                labels = torch.tensor(
                    train_labels[batch_count : batch_count + self.config["batch_size"]],
                    device=self.device,
                )
                batch_count += self.config["batch_size"]
                loss = loss_function(output, labels)
                loss.backward()
                batch_loss += loss.item()
                batches_processed += 1

                optimizer.step()
                optimizer.zero_grad()

                training_loop.set_postfix({"loss": batch_loss / batches_processed})
                training_loop.set_description(f"Fine-Tuning: epoch {j+1}/{epochs}")

            if lr_scheduler is not None:
                lr_scheduler.step()

            if validation_input_data is not None:
                testing_loop = tqdm(
                    validation_dataloader, desc="Fine-Tuning Validation"
                )
                val_loss = 0.0
                count = 0.0
                validation_batch_count = 0
                for validation_data in testing_loop:
                    batch_sentences, mask, idxs = (
                        validation_data[0],
                        validation_data[1],
                        validation_data[2],
                    )
                    batch_sentences = batch_sentences.permute(1, 0)
                    if self.config["multi_gpu"]:
                        batch_sentences = self.model.module.pe_embedding(
                            batch_sentences.long()
                        )
                    else:
                        batch_sentences = self.model.pe_embedding(
                            batch_sentences.long()
                        )
                    batch_sentences = torch.nn.functional.normalize(
                        batch_sentences, dim=2
                    )  # normalize token outputs
                    output = self._forward(batch_sentences, mask=mask)
                    val_labels = torch.tensor(
                        validation_labels[
                            validation_batch_count : validation_batch_count
                            + self.config["batch_size"]
                        ],
                        device=self.device,
                    )
                    validation_batch_count += self.config["batch_size"]
                    val_loss += loss_function(output, val_labels).item()
                    count += 1.0
                    testing_loop.set_postfix({"val_loss": val_loss / count})
        logger.info(f"Fine-Tuning Complete. Epochs: {epochs}")
        self.model.eval()
        self.fine_tuning_head.eval()

    def get_outputs(self, dataset: UCEDataset) -> np.ndarray:
        """
        Get the outputs of the fine-tuned model on a dataset.

        Parameters
        ----------
        dataset : UCEDataset
            The dataset to get the outputs for. This is the dataset returned from the `process_data` method.

        Returns
        -------
        np.ndarray
            The outputs of the model as a numpy array.
        """
        self.to(self.device)

        batch_size = self.config["batch_size"]
        dataloader = DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=False,
            collate_fn=dataset.collator_fn,
            num_workers=0,
        )

        if self.accelerator is not None:
            dataloader = self.accelerator.prepare(dataloader)

        self.model.eval()
        self.fine_tuning_head.eval()

        testing_loop = tqdm(dataloader, desc="Fine-Tuning Validation")
        outputs = []
        for validation_data in testing_loop:
            batch_sentences, mask, idxs = (
                validation_data[0],
                validation_data[1],
                validation_data[2],
            )
            batch_sentences = batch_sentences.permute(1, 0)
            if self.config["multi_gpu"]:
                batch_sentences = self.model.module.pe_embedding(batch_sentences.long())
            else:
                batch_sentences = self.model.pe_embedding(batch_sentences.long())
            batch_sentences = torch.nn.functional.normalize(
                batch_sentences, dim=2
            )  # normalize token outputs
            output = self._forward(batch_sentences, mask=mask)
            outputs.append(output.detach().cpu().numpy())

        return np.vstack(outputs)



import logging
import numpy as np
import torch


class BatchGeneTokenizer:
    def __init__(self, gene_vocab):
        self.gene_vocab = gene_vocab
        self.unknown_token = gene_vocab["unknown"]
        self.gene_map = np.vectorize(self.gene_map)

    def gene_map(self, x):
        return self.gene_vocab.get(x, self.unknown_token)

    def __call__(self, gene_names):
        toks = torch.tensor(self.gene_map(gene_names))
        unknown_mask = toks == self.unknown_token
        if unknown_mask.any():
            n_unknown = unknown_mask.sum().item()
            logging.warning(f"Warning: {n_unknown} genes not found in gene vocab")
        return toks


class BatchObsTokenizer:
    def __init__(self, vocab):
        self.vocab = vocab

    def __call__(self, obs):
        return torch.tensor(
            [
                self.vocab[key].get(obs[key], self.vocab[key]["unknown"])
                for key in self.vocab
            ],
            dtype=torch.int64,
        )
